# pylint: disable=unused-wildcard-import, wildcard-import, too-many-lines

"""
  This module solves the receiver system or the single tube
  structural problem.
"""

from abc import ABC, abstractmethod

import copy

import scipy.sparse.linalg as spla
import numpy.linalg as la

import numpy as np

from skfem import *
from skfem import mesh, element
from skfem import helpers, utils
from skfem.element.discrete_field import DiscreteField

from neml import block

from srlife import solverparams


class TubeSolver(ABC):
    """
    This class takes as input:
      1) A state object containing whatever the solver needs to execute the solve
         These objects must contain a method for dumping the required results fields into
         a Tube object
      2) A input top displacement

    It must return:
      1) A copied and updated state
      2) The force on the top face
      3) The derivative of the force with respect to the displacement

      These three are grouped in a State object
    """

    @abstractmethod
    def solve(self, tube, i, state_n, dtop):
        """
        Solve the structural tube problem for a single time step

        Parameters:
          tube:       tube object with all bcs
          i:          time index to reference in tube results
          state:      state object
          dtop:       top displacement
        """
        return

    @abstractmethod
    def setup_tube(self, tube):
        """
        Setup all the quadrature results fields in the tube

        Parameters:
          tube:       tube object
        """
        return

    @abstractmethod
    def init_state(self, tube, mat):
        """
        Initialize the solver state

        Parameters:
          tube:       tube object
          mat:        NEML material
        """
        return

    @abstractmethod
    def dump_state(self, tube, i, state):
        """
        Update the required results fields in a tube object with the
        current state

        Parameters:
          tube:       tube to update
          i:          which time step this is
          state:      state object
        """
        return


def mesh_tube(tube):
    """
    Make a simple, regular Cartesian mesh of a Tube using:
      1D:   linear Lagrange (line)
      2D:   bilinear Lagrange (Quad4)
      3D:   trilinear Lagrange (Hex8)
    elements
    """
    if tube.abstraction == "3D":
        return mesh3D(tube)
    elif tube.abstraction == "2D":
        return mesh2D(tube)
    elif tube.abstraction == "1D":
        return mesh1D(tube)
    else:
        raise ValueError("Unknown tube abstraction %s!" % tube.abstraction)


def mesh1D(tube):
    """
    Make a 1D linear Lagrange mesh of a tube
    """
    return mesh.MeshLine(tube.mesh[0].flatten())


def mesh2D(tube):
    """
    Make a 2D Cartesian Lagrange mesh of a tube
    """
    rs = np.linspace(tube.r - tube.t, tube.r, tube.nr)
    ts = np.linspace(0, 2 * np.pi, tube.nt + 1)[: tube.nt]

    coords = np.swapaxes(
        np.array([[ri * np.cos(ts), ri * np.sin(ts)] for ri in rs]), 0, 1
    ).reshape(2, len(rs) * len(ts))

    conn = []
    for i in range(tube.nr - 1):
        for j in range(tube.nt):
            conn.append(
                [
                    i * tube.nt + j,
                    i * tube.nt + ((j + 1) % tube.nt),
                    (i + 1) * tube.nt + ((j + 1) % tube.nt),
                    (i + 1) * tube.nt + j,
                ]
            )

    conn = np.array(conn, dtype=int).T

    return mesh.MeshQuad(coords, conn)


def mesh3D(tube):
    """
    Make a 3D Cartesian Lagrange mesh of a tube
    """
    rs = np.linspace(tube.r - tube.t, tube.r, tube.nr)
    ts = np.linspace(0, 2 * np.pi, tube.nt + 1)[: tube.nt]
    zs = np.linspace(0, tube.h, tube.nz)

    npr = tube.nt * tube.nz
    npt = tube.nz

    coords = np.ascontiguousarray(
        np.array(
            [[r * np.cos(t), r * np.sin(t), z] for r in rs for t in ts for z in zs]
        ).T
    )

    mapper = lambda r, c, h: r * npr + (c % tube.nt) * npt + h

    conn = []
    for i in range(tube.nr - 1):
        for j in range(tube.nt):
            for k in range(tube.nz - 1):
                conn.append(
                    [
                        mapper(i + 1, j + 1, k),
                        mapper(i, j + 1, k),
                        mapper(i + 1, j + 1, k + 1),
                        mapper(i + 1, j, k),
                        mapper(i, j + 1, k + 1),
                        mapper(i, j, k),
                        mapper(i + 1, j, k + 1),
                        mapper(i, j, k + 1),
                    ]
                )

    conn = np.ascontiguousarray(np.array(conn, dtype=int).T)

    return mesh.MeshHex(coords, conn)


class PythonTubeSolver(TubeSolver):
    """
    Tube solver class coded up with scikit.fem and the scipy sparse solvers
    """

    def __init__(
        self,
        pset=solverparams.ParameterSet(),
        rtol=1.0e-6,
        atol=1.0e-8,
        qorder=1,
        dof_tol=1.0e-6,
        miter=10,
        verbose=False,
        max_divide=4,
        force_divide=False,
        max_linesearch=10,
    ):
        """
        Setup the solver with common parameters

        Additional Parameters:
          pset:            parameter set with solver parameters
          rtol:            relative tolerance for NR iterations
          atol:            absolute tolerance for NR iterations
          qorder:          quadrature order
          dof_to:l         geometric tolerance on finding boundary
                           degrees of freedom
          miter:           maximum newton-raphson iterations
          verbose:         verbose solve
          max_divide:      maximum adaptive integration subdivisions
          force_divide:    force adaptive substeps, for tests or
                           accuracy checks
          max_linesearch   maximum line search cutbacks
        """
        self.qorder = qorder
        self.rtol = pset.get_default("rtol", rtol)
        self.atol = pset.get_default("atol", atol)
        self.dof_tol = pset.get_default("dof_tol", dof_tol)
        self.miter = pset.get_default("miter", miter)
        self.verbose = pset.get_default("verbose", verbose)
        self.max_divide = pset.get_default("max_divide", max_divide)
        self.force_divide = pset.get_default("force_divide", force_divide)
        self.max_search = pset.get_default("max_linesearch", max_linesearch)
        self.solver_options = {
            "rtol": self.rtol,
            "atol": self.atol,
            "dof_tol": self.dof_tol,
            "miter": self.miter,
            "verbose": self.verbose,
            "max_linesearch": self.max_search,
        }

    def setup_tube(self, tube):
        """
        Add all the quadrature result fields we will need

        Parameters:
          tube        tube object
        """
        suffixes = ["_xx", "_yy", "_zz", "_yz", "_xz", "_xy"]
        fields = ["stress", "strain", "mechanical_strain", "thermal_strain"]

        for field in fields:
            for suffix in suffixes:
                tube.add_blank_quadrature_results(
                    field + suffix, (tube.ntime,) + self.qshape(tube)
                )

        tube.add_blank_quadrature_results(
            "temperature", (tube.ntime,) + self.qshape(tube)
        )

        suffixes = ["_x", "_y", "_z"]
        for i in range(tube.ndim):
            tube.add_blank_results(
                "disp" + suffixes[i], (tube.ntime,) + tube.dim[: tube.ndim]
            )

    def qshape(self, tube):
        """
        Number of elements x number of quadrature points
        """
        if tube.dim[2] == 1:
            nz = 1
        else:
            nz = tube.dim[2] - 1
        nelem = (tube.dim[0] - 1) * (tube.dim[1]) * nz
        return (nelem, (self.qorder + 1) ** tube.ndim)

    def _setup_state(self, sf, tube, i, state_n):
        """
        Setup all the information needed to solve the problem for
        some step fraction of a whole step

        Parameters:
          sf:         step fraction
          tube:       tube object with all the BC information
          i:          time index for state n+1
          state_n:    state at time n
        """
        state_next = state_n.copy()

        t = tube.times[i - 1] + (tube.times[i] - tube.times[i - 1]) * sf

        if "temperature" in tube.results:
            T_np1 = self._res2quad(
                state_next, self._tube2fea(tube, tube.results["temperature"][i])
            )
            T_n = self._res2quad(
                state_n, self._tube2fea(tube, tube.results["temperature"][i - 1])
            )
            T = T_n + (T_np1 - T_n) * sf
            state_next.temperature = T
        else:
            state_next.temperature = np.zeros(state_next.temperature.shape)

        if tube.pressure_bc:
            p = tube.pressure_bc.pressure(t)
        else:
            p = 0

        return state_next, p, t

    # pylint: disable=too-many-branches
    def solve(self, tube, i, state_n, dtop):
        """
        Solve the structural tube problem for a single time step
        using adaptive integration

        Parameters:
          tube:       tube object with all bcs
          i:          time index to reference in tube results
          state:      state object
          dtop:       top displacement
        """
        # Setup initial state_n
        if "temperature" in tube.results:
            state_n.temperature = self._res2quad(
                state_n, self._tube2fea(tube, tube.results["temperature"][i - 1])
            )
        else:
            state_n.temperature = np.zeros(state_n.temperature.shape)

        state_last = state_n.copy()

        t_last = tube.times[i - 1]
        if tube.pressure_bc:
            p_last = tube.pressure_bc.pressure(t_last)
        else:
            p_last = 0.0

        # Adaptively integrate
        tprog = 2**self.max_divide
        cprog = 0
        if self.force_divide:
            inc = 1
            mdiv = self.max_divide - 1
        else:
            inc = tprog
            mdiv = 0

        while cprog < tprog:
            sf = float(cprog + inc) / float(tprog)

            state_next, p_next, t_next = self._setup_state(sf, tube, i, state_n)

            try:
                if tube.ndim == 1:
                    solve_python_1d(
                        state_last,
                        t_last,
                        p_last,
                        state_next,
                        t_next,
                        p_next,
                        dtop * sf,
                        self.solver_options,
                    )
                elif tube.ndim == 2:
                    solve_python_2d(
                        state_last,
                        t_last,
                        p_last,
                        state_next,
                        t_next,
                        p_next,
                        dtop * sf,
                        self.solver_options,
                    )
                elif tube.ndim == 3:
                    solve_python_3d(
                        state_last,
                        t_last,
                        p_last,
                        state_next,
                        t_next,
                        p_next,
                        dtop * sf,
                        self.solver_options,
                    )
                else:
                    raise ValueError("Unknown dimension %i" % tube.ndim)
            except RuntimeError:
                inc /= 2
                mdiv += 1
                if mdiv >= self.max_divide:
                    break

            state_last = state_next
            t_last = t_next
            p_last = p_next
            cprog += inc

        if mdiv >= self.max_divide:
            raise RuntimeError("Adaptive integration failed")

        return state_next

    def init_state(self, tube, mat, i=None):
        """
        Initialize the solver state

        Parameters:
          tube:       tube object
          mat:        NEML material

        Additional Parameters:
          i:          if not none, also dump state into the tube
        """
        return PythonTubeSolver.State(tube, mat, self.qorder, i=i)

    def dump_state(self, tube, i, state):
        """
        Update the required results fields in a tube object with the
        current state

          Parameters:
          tube:       tube to update
          i:          which time step this is
          state:      state object
        """
        # Displacement
        suffixes = ["_x", "_y", "_z"]
        for k in range(state.ndim):
            tube.results["disp" + suffixes[k]][i] = self._fea2tube(
                tube, state.displacements[k :: state.ndim]
            )

        # The tensor fields
        order = ["_xx", "_yy", "_zz", "_yz", "_xz", "_xy"]
        inds = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]
        fields = ["stress", "strain", "mechanical_strain", "thermal_strain"]
        data = [
            state.stress,
            state.strain,
            state.mechanical_strain,
            state.thermal_strain,
        ]
        for f, d in zip(fields, data):
            for ind, o in zip(inds, order):
                tube.quadrature_results[f + o][i] = self._fea2tube_element(tube, d[ind])

        tube.quadrature_results["temperature"][i] = self._fea2tube_element(
            tube, state.temperature
        )

    def _fea2tube_element(self, tube, f):
        """
        Rearrange the elements (first index) of the field to match the vtk
        convention
        """
        return f

    def _tube2fea(self, tube, f):
        """
        Map a result field in the tube to the flat FEA vector
        """
        return f.flatten()

    def _fea2tube(self, tube, f):
        """
        Map a result field in the FEA to the right shape for the tube
        """
        return f.reshape(tube.dim[: tube.ndim])

    def _quad2res(self, state, f):
        """
        Quadrature results to a field result

        This does Laplacian smoothing
        """
        Md = asm(BilinearForm(lambda u, v, w: v * u), state.sbasis)
        fd = asm(
            LinearForm(lambda v, w: v * w["values"]),
            state.sbasis,
            values=DiscreteField(f),
        )

        return spla.spsolve(Md, fd)

    def _res2quad(self, state, f):
        """
        Results field to the quadrature points
        """
        return state.sbasis.interpolate(f).value

    class State:
        """
        Subclass for maintaining state with the python solver
        """

        def __init__(self, tube, mat, qorder, i=None):
            """
            Initialize a full state object
            """
            self.mat = mat
            self.mesh = mesh_tube(tube)
            self.qorder = qorder
            self.ndim = tube.ndim

            self.ri = tube.r - tube.t
            self.ro = tube.r
            self.h = tube.h

            # Define the pressure boundary
            self.define_boundary(tube)

            # Base element type
            self.betype = (
                element.ElementLineP1(),
                element.ElementQuad1(),
                element.ElementHex1(),
            )[tube.ndim - 1]

            # Define the scalar basis for interpolation
            self.define_scalar_basis()

            # Define the interior basis for momentum balance
            self.define_interior_basis()

            # Define the side basis for pressure
            self.define_exterior_basis()

            # Now that is out of the way, setup the actual required storage
            self.stress = np.zeros((3, 3, self.basis.nelems, self.nqi))
            self.strain = np.zeros((3, 3, self.basis.nelems, self.nqi))
            self.mechanical_strain = np.zeros((3, 3, self.basis.nelems, self.nqi))
            self.thermal_strain = np.zeros((3, 3, self.basis.nelems, self.nqi))
            self.history = np.repeat(
                self.material.init_store()[:, np.newaxis], self.nq, axis=1
            ).reshape(self.material.nstore, self.basis.nelems, self.nqi)
            self.tangent = np.zeros((3, 3, 3, 3, self.basis.nelems, self.nqi))
            self.temperature = np.zeros((self.basis.nelems, self.nqi))
            self.displacements = self.basis.zeros()
            self.time = 0.0

            self.energy = np.zeros((self.basis.nelems, self.nqi))
            self.dissipation = np.zeros((self.basis.nelems, self.nqi))

            self.force = 0.0
            self.stiffness = 0.0

            if i is not None:
                self.temperature = self.sbasis.interpolate(
                    tube.results["temperature"][i].flatten()
                ).value

        @property
        def material(self):
            """
            Getter for material object
            """
            return get_mat(self.mat)

        def define_boundary(self, tube, tol=0.5):
            """
            Define the pressure boundary

            Parameters:
              tube:     tube object for geometry

            Additional parameters:
              tol:      thickness tolerance for finding faces
            """
            atol = tol * tube.t
            if self.ndim == 1:
                self.mesh = self.mesh.with_boundaries(
                    {
                        "pressure": lambda x: np.logical_and(
                            x > tube.r - tube.t - atol, x < tube.r - tube.t + atol
                        )
                    }
                )
            else:
                self.mesh = self.mesh.with_boundaries(
                    {
                        "pressure": lambda x: np.logical_and(
                            np.sqrt(x[0] ** 2.0 + x[1] ** 2.0) > tube.r - tube.t - atol,
                            np.sqrt(x[0] ** 2.0 + x[1] ** 2.0) < tube.r - tube.t + atol,
                        )
                    }
                )

        def define_scalar_basis(self):
            """
            Define the scalar basis for mapping around result fields
            """
            self.sbasis = InteriorBasis(
                mesh=self.mesh, elem=self.betype, intorder=self.qorder
            )

        def define_interior_basis(self):
            """
            Define the interior basis for the balance equation
            """
            etype = element.ElementVectorH1(self.betype)
            self.basis = InteriorBasis(mesh=self.mesh, elem=etype, intorder=self.qorder)

        def define_exterior_basis(self):
            """
            Define the sideset basis for applying pressure
            """
            etype = element.ElementVectorH1(self.betype)
            self.pbasis = FacetBasis(
                mesh=self.mesh,
                elem=etype,
                intorder=self.qorder,
                facets=self.mesh.boundaries["pressure"],
            )

        def copy(self):
            """
            Return a copy

            Soft copy the scikit-fem stuff
            Hard copy the results
            """
            new = copy.copy(self)
            new.stress = np.copy(self.stress)
            new.strain = np.copy(self.strain)
            new.mechanical_strain = np.copy(self.mechanical_strain)
            new.thermal_strain = np.copy(self.thermal_strain)
            new.history = np.copy(self.history)
            new.tangent = np.copy(self.tangent)
            new.temperature = np.copy(self.temperature)
            new.displacements = np.copy(self.displacements)
            new.energy = np.copy(self.energy)
            new.dissipation = np.copy(self.dissipation)
            return new

        @property
        def nqi(self):
            """
            Integration points per element
            """
            return len(self.basis.quadrature[1])

        @property
        def nq(self):
            """
            Total number of integration points (size of array)
            """
            return self.nqi * self.basis.mesh.nelements

        @property
        def ne(self):
            """
            Number of elements
            """
            return self.basis.mesh.nelements


def solve_python_1d(
    state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1, solver_options
):
    """
    Solve an increment using the python 1d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
    """
    solver = PythonSolver(
        state_n, state_np1, solver_options, set_strain=top_np1 / state_np1.h
    )

    solver.solve(t_n, t_np1, p_np1)


def solve_python_2d(
    state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1, solver_options
):
    """
    Solve an increment using the python 2d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
    """
    solver = PythonSolver(
        state_n, state_np1, solver_options, set_strain=top_np1 / state_np1.h
    )

    # Fix the xy line
    node = state_np1.mesh.nodes_satisfying(
        lambda x: np.logical_and(
            np.logical_and(
                x[0] < state_np1.ro + solver_options["dof_tol"],
                x[0] > state_np1.ro - solver_options["dof_tol"],
            ),
            np.logical_and(
                x[1] < solver_options["dof_tol"], x[1] > -solver_options["dof_tol"]
            ),
        )
    )
    dofs_x = state_np1.basis.nodal_dofs[0, node].flatten()
    solver.add_dirichlet_bc(dofs_x, 0.0)
    dofs_y = state_np1.basis.nodal_dofs[1, node].flatten()
    solver.add_dirichlet_bc(dofs_y, 0.0)

    # Fix the y line
    node = state_np1.mesh.nodes_satisfying(
        lambda x: np.logical_and(
            np.logical_and(
                -x[0] < state_np1.ro + solver_options["dof_tol"],
                -x[0] > state_np1.ro - solver_options["dof_tol"],
            ),
            np.logical_and(
                x[1] < solver_options["dof_tol"], x[1] > -solver_options["dof_tol"]
            ),
        )
    )
    dofs_y = state_np1.basis.nodal_dofs[1, node].flatten()
    solver.add_dirichlet_bc(dofs_y, 0.0)

    solver.solve(t_n, t_np1, p_np1)


def solve_python_3d(
    state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1, solver_options
):
    """
    Solve an increment using the python 3d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
    """
    solver = PythonSolver(state_n, state_np1, solver_options)

    # Fix the xy line
    node = state_np1.mesh.nodes_satisfying(
        lambda x: np.logical_and(
            np.logical_and(
                np.logical_and(
                    x[0] < state_np1.ro + solver_options["dof_tol"],
                    x[0] > state_np1.ro - solver_options["dof_tol"],
                ),
                np.logical_and(
                    x[1] < solver_options["dof_tol"], x[1] > -solver_options["dof_tol"]
                ),
            ),
            x[2] < solver_options["dof_tol"],
        )
    )
    dofs_x = state_np1.basis.nodal_dofs[0, node].flatten()
    solver.add_dirichlet_bc(dofs_x, 0.0)
    dofs_y = state_np1.basis.nodal_dofs[1, node].flatten()
    solver.add_dirichlet_bc(dofs_y, 0.0)

    # Fix the y line
    node = state_np1.mesh.nodes_satisfying(
        lambda x: np.logical_and(
            np.logical_and(
                np.logical_and(
                    -x[0] < state_np1.ro + solver_options["dof_tol"],
                    -x[0] > state_np1.ro - solver_options["dof_tol"],
                ),
                np.logical_and(
                    x[1] < solver_options["dof_tol"], x[1] > -solver_options["dof_tol"]
                ),
            ),
            x[2] < solver_options["dof_tol"],
        )
    )
    dofs_y = state_np1.basis.nodal_dofs[1, node].flatten()
    solver.add_dirichlet_bc(dofs_y, 0.0)

    # Fix the bottom face
    node = state_np1.mesh.nodes_satisfying(lambda x: x[2] < solver_options["dof_tol"])
    dofs = state_np1.basis.nodal_dofs[2, node].flatten()
    solver.add_dirichlet_bc(dofs, 0.0)

    # Fix the top face
    node = state_np1.mesh.nodes_satisfying(
        lambda x: x[2] > state_np1.h - solver_options["dof_tol"]
    )
    dofs = state_np1.basis.nodal_dofs[2, node].flatten()
    solver.add_dirichlet_bc(dofs, top_np1)

    solver.solve(t_n, t_np1, p_np1)


class PythonSolver:
    """
    Actually manages the solve with scikit-fem
    """

    def __init__(self, state_n, state_np1, solver_options, set_strain=None):
        """
        Setup solver

        Parameters:
          state_n         previous state
          state_np1       current state
          solver_options  various 'how to solve' parameters

        Additional parameters:
          set_strain      set the zz strain to this value, if given
        """
        self.state_n = state_n
        self.state_np1 = state_np1
        self.options = solver_options
        self.set_strain = set_strain

        self.ebcs = []

        self.edofs = []
        self.evalues = []

        # Initialize the guess
        self.setup_guess()

        # The operators!
        if self.ndim == 1:
            # 1D axisymmetric is different than 2D/3D cartesian
            self.internal = LinearForm(
                lambda v, w: (
                    v.grad[0][0] * w["radial"]
                    - v.value[0] * w["radial"] / w.x[0]
                    + v.value[0] * w["hoop"] / w.x[0]
                )
            )
            self.jac = BilinearForm(
                lambda u, v, w: v.grad[0][0] * w["Crr"] * u.grad[0][0]
                + v.grad[0][0] * w["Crt"] * u.value[0] / w.x[0]
                - v.value[0]
                * (w["Crr"] * u.grad[0][0] + w["Crt"] * u.value[0] / w.x[0])
                / w.x[0]
                + v.value[0]
                * (w["Ctr"] * u.grad[0][0] + w["Ctt"] * u.value[0] / w.x[0])
                / w.x[0]
            )
        else:
            self.internal = LinearForm(
                lambda v, w: helpers.ddot(helpers.sym_grad(v), w["stress"])
            )
            self.jac = BilinearForm(
                lambda u, v, w: helpers.ddot(
                    helpers.sym_grad(v), helpers.ddot(w["C"], helpers.sym_grad(u))
                )
            )
        self.external = LinearForm(
            lambda v, w: helpers.dot((-1.0) * w["pressure"] * w.n, v)
        )

    @property
    def ndim(self):
        """
        Problem dimension
        """
        return self.state_np1.ndim

    def setup_guess(self):
        """
        Estimate the displacements
        """
        self.state_np1.displacements = self.state_n.displacements

    def solve(self, t_n, t_np1, p):
        """
        Do the actual solve using Newton-Raphson iteration

        Parameters:
          t_n     previous time
          t_np1   next time
          p       pressure
        """
        # Setup the time step
        self.state_n.time = t_n
        self.state_np1.time = t_np1

        # Assemble the dirichlet BCs into big vectors
        self.assemble_dirichlet()

        # Do the initial stress update
        self.update_state()

        # Calculate the initial residual and jacobian
        R = self.residual(p)
        nR0 = la.norm(R[self.kdofs])
        nR = nR0

        # Printing, if you want
        if self.options["verbose"]:
            print("Iter\tnR\t\tnR/nR0\t\talpha")
            print("%i\t%3.2e\t" % (0, nR0))

        # Newton-Raphson iteration
        for i in range(self.options["miter"]):
            # Can immediately skip if the initial residual is
            # less than the absolute tol
            if nR0 < self.options["atol"]:
                if self.options["verbose"]:
                    print("")
                break
            # Get the actual sparse jacobian
            J = self.jacobian()
            # Direction
            dx = self.linear_solve(J, R)
            # Linesearch
            alpha = 1.0
            x0 = np.copy(self.state_np1.displacements)
            nR_start = nR
            for _ in range(self.options["max_linesearch"]):
                # Trial
                self.state_np1.displacements = x0 - alpha * dx
                # Recalculate the state
                self.update_state()
                # Calculate the residual
                R = self.residual(p)
                # Norm of residual
                nR = la.norm(R[self.kdofs])
                # Check linesearch criteria
                if nR < nR_start:
                    break
                alpha /= 2.0

            # Check convergence
            if self.options["verbose"]:
                print("%i\t%3.2e\t%3.2e\t%3.2e" % (i + 1, nR, nR / nR0, alpha))
            if nR < self.options["atol"] or nR / nR0 < self.options["rtol"]:
                if self.options["verbose"]:
                    print("")
                break
        else:
            raise RuntimeError("Nonlinear iteration did not converge!")

        # Calculate the axial force and stiffness
        J = self.jacobian()  # Need the current tangent
        if self.state_np1.ndim == 1 or self.state_np1.ndim == 2:
            self.calculate_axial_from_stress(R, J)
        else:
            self.calculate_axial_from_fea(R, J)

    def calculate_axial_from_stress(self, R, J):
        """
        Calculate the axial force by integrating sigma_zz over the area

        Parameters:
          R       final residual
          J       final jacobian
        """
        # pylint: disable=no-value-for-parameter, invalid-unary-operand-type
        if self.state_np1.ndim == 1:
            dx = (
                self.state_np1.basis.interpolate(self.state_np1.mesh.p[0]).value[0]
                * self.state_np1.basis.dx
                * 2.0
                * np.pi
            )
        else:
            dx = self.state_np1.basis.dx

        fake_force = self._internal_force(self.state_np1.tangent[:, :, 2, 2])
        de = utils.solve(
            *utils.condense(J, fake_force, D=self.edofs),
            solver=utils.solver_direct_scipy()
        )
        fake_strain = self.calculate_strain(de)
        integrand1 = np.einsum("iijk", self.state_np1.tangent[2, 2] * fake_strain)
        integrand2 = self.state_np1.tangent[2, 2, 2, 2]

        self.state_np1.force = np.sum(self.state_np1.stress[2, 2] * dx)
        self.state_np1.stiffness = (
            np.sum((-integrand1 + integrand2) * dx) / self.state_np1.h
        )

    def calculate_axial_from_fea(self, R, J):
        """
        Calculate the axial force by summing the discrete forces on the top dofs

        Parameters:
          R       final residual
          J       final jacobian
        """
        node = self.state_np1.mesh.nodes_satisfying(
            lambda x: x[2] > self.state_np1.h - self.options["dof_tol"]
        )
        dofs = self.state_np1.basis.nodal_dofs[2, node].flatten()

        # There is a better way to do this
        sdofs = [list(self.edofs).index(d) for d in dofs]
        dotme = np.zeros((len(self.edofs),))
        dotme[sdofs] = 1.0

        K = self.state_np1.basis.complement_dofs(self.edofs)

        # Straightforward
        self.state_np1.force = np.sum(R[dofs])

        # Expensive but eh can optimize later
        Jl = J.tolil()
        J11 = Jl[K, :][:, K].tocsr()
        J12 = Jl[K, :][:, self.edofs]

        J21 = Jl[self.edofs, :][:, K]
        J22 = Jl[self.edofs, :][:, self.edofs]

        self.state_np1.stiffness = np.dot(
            dotme, J22.dot(dotme) - J21.dot(spla.spsolve(J11, J12.dot(dotme)))
        )

    def linear_solve(self, J, R):
        """
        Do the linear solve for a particular jacobian and residual

        Parameters:
          J       Jacobian sparse matrix
          R       residual vector
        """
        # pylint: disable=no-value-for-parameter
        return utils.solve(
            *utils.condense(J, R, D=self.edofs), solver=utils.solver_direct_scipy()
        )

    def residual(self, p):
        """
        Actually assemble the residual

        Parameters:
          p       Pressure, for the BC
        """
        F_int = self._internal_force(self.state_np1.stress)
        pv = self.state_np1.pbasis.interpolate(self.state_np1.pbasis.zeros() + p)
        F_ext = asm(self.external, self.state_np1.pbasis, pressure=pv)

        return F_int - F_ext

    def _internal_force(self, stress):
        """
        Assemble the internal force

        Parameters:
          stress      stresses to use
        """
        if self.ndim == 1:
            return self._internal_1d(stress[0, 0], stress[1, 1])
        else:
            return self._internal_nd(stress[: self.ndim, : self.ndim])

    def _internal_1d(self, radial, hoop):
        """
        Assemble the internal force for the axisymmetric case

        Parameters:
          radial      radial stress
          hoop        hoop stress
        """
        return asm(
            self.internal,
            self.state_np1.basis,
            radial=DiscreteField(radial),
            hoop=DiscreteField(hoop),
        )

    def _internal_nd(self, stress):
        """
        Internal force of the Cartesian cases

        Parameters:
          stress      stresses to use
        """
        return asm(self.internal, self.state_np1.basis, stress=DiscreteField(stress))

    def jacobian(self):
        """
        Actually assemble the jacobian
        """
        if self.ndim == 1:
            return asm(
                self.jac,
                self.state_np1.basis,
                Crr=DiscreteField(self.state_np1.tangent[0, 0, 0, 0]),
                Crt=DiscreteField(self.state_np1.tangent[0, 0, 1, 1]),
                Ctt=DiscreteField(self.state_np1.tangent[1, 1, 1, 1]),
                Ctr=DiscreteField(self.state_np1.tangent[1, 1, 0, 0]),
            )
        else:
            return asm(
                self.jac,
                self.state_np1.basis,
                C=DiscreteField(
                    self.state_np1.tangent[
                        : self.ndim, : self.ndim, : self.ndim, : self.ndim
                    ]
                ),
            )

    def add_dirichlet_bc(self, dofs, value):
        """
        Add a dirichlet BC

        Parameters:
          dofs        plain dof list
          value       value to set
        """
        self.ebcs.append((dofs, value))

    def assemble_dirichlet(self):
        """
        Assemble the dirichlet BCs into big vectors
        """
        self.edofs = []
        self.evalues = []

        for d, v in self.ebcs:
            self.edofs.extend(list(d))
            self.evalues.extend([v] * len(d))

        self.edofs = np.array(self.edofs, dtype=int)
        self.evalues = np.array(self.evalues)

        inds = np.argsort(self.edofs)
        self.edofs = self.edofs[inds]
        self.evalues = self.evalues[inds]
        self.kdofs = self.state_np1.basis.complement_dofs(self.edofs)

    def update_state(self):
        """
        Make the state consistent with the new displacements
        """
        # Enforce the Dirichlet BCs in the vector
        self.state_np1.displacements[self.edofs] = self.evalues

        # Calculate total strain
        self.state_np1.strain = self.calculate_strain(self.state_np1.displacements)

        # Add the optional extra strain
        if self.set_strain is not None:
            self.state_np1.strain[2, 2] = self.set_strain

        # Calculate thermal strain and mechanical strain
        self.calculate_mechanical_strain()

        # Calculate stress and tangent
        self.calculate_stress_update()

    def calculate_strain(self, D):
        """
        Calculate the strain based on the current displacements

        Parameters:
          D           displacements to use
        """
        U = self.state_np1.basis.interpolate(D)

        E = np.zeros((3, 3) + U.value[0].shape)

        E[: self.ndim, : self.ndim] = helpers.sym_grad(U)
        if self.ndim == 1:
            # Cast needed b/c they don't have a divide operator!
            E[1, 1] = np.array(U) / np.array(
                self.state_np1.basis.interpolate(self.state_np1.mesh.p.flatten())
            )

        return E

    def calculate_mechanical_strain(self):
        """
        Calculate the thermal strain increment and update mechanical strain
        """
        # Grab CTE values from NEML
        cte_old = np.zeros(self.state_n.temperature.shape)
        cte_new = np.zeros(self.state_np1.temperature.shape)
        for ind in np.ndindex(self.state_np1.temperature.shape):
            cte_old[ind] = self.state_n.material.alpha(self.state_n.temperature[ind])
            cte_new[ind] = self.state_np1.material.alpha(
                self.state_np1.temperature[ind]
            )

        # Thermal strain = thermal_strain_old + alpha * Tdot
        self.state_np1.thermal_strain = self.state_n.thermal_strain + np.eye(3)[
            :, :, np.newaxis, np.newaxis
        ] * (
            (cte_new + cte_old)
            / 2.0
            * (self.state_np1.temperature - self.state_n.temperature)
        )

        # Mechanical strain = total - thermal
        self.state_np1.mechanical_strain = (
            self.state_np1.strain - self.state_np1.thermal_strain
        )

    def calculate_stress_update(self):
        """
        The expensive function: update the stress and tangent
        """
        # Ya, you can't a view of these arrays...
        # We could reorder everything and then it would just work natively
        stress_np1 = np.zeros((self.state_np1.nq, 3, 3))
        hist_np1 = np.zeros((self.state_np1.nq, self.state_np1.material.nstore))
        A_np1 = np.zeros((self.state_np1.nq, 3, 3, 3, 3))

        # pylint: disable=c-extension-no-member
        block.block_evaluate(
            self.state_np1.material,
            self.state_np1.mechanical_strain.transpose(2, 3, 0, 1).reshape(-1, 3, 3),
            self.state_n.mechanical_strain.transpose(2, 3, 0, 1).reshape(-1, 3, 3),
            self.state_np1.temperature.flatten(),
            self.state_n.temperature.flatten(),
            self.state_np1.time,
            self.state_n.time,
            stress_np1,
            self.state_n.stress.transpose(2, 3, 0, 1).reshape(-1, 3, 3),
            hist_np1,
            self.state_n.history.transpose(1, 2, 0).reshape(
                -1, self.state_n.material.nstore
            ),
            A_np1,
            self.state_np1.energy.flatten(),
            self.state_n.energy.flatten(),
            self.state_np1.dissipation.flatten(),
            self.state_n.dissipation.flatten(),
        )

        self.state_np1.stress = stress_np1.reshape(
            (self.state_np1.ne, self.state_np1.nqi, 3, 3)
        ).transpose(2, 3, 0, 1)
        self.state_np1.history = hist_np1.reshape(
            (self.state_np1.ne, self.state_np1.nqi, self.state_np1.material.nstore)
        ).transpose(2, 0, 1)
        self.state_np1.tangent = A_np1.reshape(
            (self.state_np1.ne, self.state_np1.nqi, 3, 3, 3, 3)
        ).transpose(2, 3, 4, 5, 0, 1)


def get_mat(thing):
    """
    Small helper to wrap NEML for pickling issues
    """
    try:
        return thing.get_neml_model()
    except AttributeError:
        return thing



# pylint: disable=dangerous-default-value, unused-import
"""
  This module contains material models containing thermal, fluid, and
  material properties.  These models can be stored to and recalled from
  XML files for archiving.
"""

from collections import ChainMap
import xml.etree.ElementTree as ET

import numpy as np
import scipy.interpolate as inter

from neml import parse, models


class DeformationMaterial:
    """
    Incredibly thin wrapper around a NEML XML file
    """

    def __init__(self, xmlfile, modelname):
        """
        Parameters:
          xmlfile:    file location for the input file
          modelname:  which model to load
        """
        self.xmlfile = xmlfile
        self.modelname = modelname

    def get_neml_model(self):
        """
        Return the actual model for use in a solve
        """
        return parse.parse_xml(self.xmlfile, self.modelname)


class ThermalMaterial:
    """
    Material thermal properties.

    This object needs to provide:
      1) material name
      2) the conductivity, as a function of temperature and its derivative
      3) the diffusivity, as a function of temperature and its derivative
    """

    def get_dict(self):
        """
        Returns the data as a dictionary
        """
        raise NotImplementedError()

    def get_type(self):
        """
        Return the string type for the data
        """
        raise NotImplementedError()

    @classmethod
    def load(cls, fname, modelname):
        """
        Load from a dictionary

        Parameters:
          fname       filename
        """
        tag, typ = find_name(fname, modelname)
        data = load_node(tag)[modelname]

        if typ == "PiecewiseLinearThermalMaterial":
            return PiecewiseLinearThermalMaterial.load(data)
        elif typ == "ConstantThermalMaterial":
            return ConstantThermalMaterial.load(data)
        else:
            raise ValueError("Unknown ThermalMaterial type %s" % typ)

    def save(self, fname, modelname):
        """
        Save the model to an XML file as modelname

        Parameters:
          fname       filename to use
          modelname   (base tag) to use
        """
        root = ET.Element("models")

        save_node(modelname, self.get_dict(), root, attrib={"type": self.get_type()})

        tree = ET.ElementTree(element=root)
        tree.write(fname)


class PiecewiseLinearThermalMaterial(ThermalMaterial):
    """
    Interpolate thermal properties linearly from a table
    """

    def __init__(self, name, temps, cond, diff):
        """
        Properties:
          name:           material name
          temps:          list of temperature points
          cond:           list of conductivity values
          diff:           list of diffusivity values
        """
        if len(temps) != len(cond) or len(temps) != len(diff):
            raise ValueError(
                "The lists of temperatures, conductivity,"
                "and diffusivity values must have equal lengths!"
            )

        self.name = name
        self.temps = np.array(temps)
        self.cond = np.array(cond)
        self.diff = np.array(diff)

        self.fcond, self.dfcond = make_piecewise(self.temps, self.cond)
        self.fdiff, self.dfdiff = make_piecewise(self.temps, self.diff)

    def conductivity(self, T):
        """
        Conductivity as a function of temperature

        Parameters:
          T       temperature
        """
        return self.fcond(T)

    def diffusivity(self, T):
        """
        Diffusivity as a function of temperature

        Parameters:
          T       temperature
        """
        return self.fdiff(T)

    def dconductivity(self, T):
        """
        Derivative of conductivity as a function of temperature

        Parameters:
          T       temperature
        """
        return self.dfcond(T)

    def ddiffusivity(self, T):
        """
        Derivative of diffusivity as a function of temperature

        Parameters:
          T       temperature
        """
        return self.dfdiff(T)

    def get_type(self):
        """
        String type
        """
        return "PiecewiseLinearThermalMaterial"

    def get_dict(self):
        """
        Pickled dictionary
        """
        return {
            "name": self.name,
            "temps": string_array(self.temps),
            "cond": string_array(self.cond),
            "diff": string_array(self.diff),
        }

    @classmethod
    def load(cls, values):
        """
        Load from a dictionary

        Parameters:
          values  dictionary values
        """
        return cls(
            values["name"],
            destring_array(values["temps"]),
            destring_array(values["cond"]),
            destring_array(values["diff"]),
        )


class ConstantThermalMaterial(ThermalMaterial):
    """
    Constant thermal properties
    """

    def __init__(self, name, k, alpha):
        """
        Properties:
          name:           material name
          k:              conductivity
          alpha:          diffusivity
        """
        self.name = name
        self.cond = k
        self.diff = alpha

    def conductivity(self, T):
        """
        Conductivity as a function of temperature

        Parameters:
          T       temperature
        """
        return T * 0.0 + self.cond

    def diffusivity(self, T):
        """
        Diffusivity as a function of temperature

        Parameters:
          T       temperature
        """
        return T * 0.0 + self.diff

    def dconductivity(self, T):
        """
        Derivative of conductivity as a function of temperature

        Parameters:
          T       temperature
        """
        return T * 0.0

    def ddiffusivity(self, T):
        """
        Derivative of diffusivity as a function of temperature

        Parameters:
          T       temperature
        """
        return T * 0.0

    def get_type(self):
        """
        String type
        """
        return "ConstantThermalMaterial"

    def get_dict(self):
        """
        Pickled dictionary
        """
        return {"name": self.name, "k": str(self.cond), "alpha": str(self.diff)}

    @classmethod
    def load(cls, values):
        """
        Load from a dictionary

        Parameters:
          values  dictionary values
        """
        return cls(values["name"], float(values["k"]), float(values["alpha"]))


class FluidMaterial:
    """
    Properties for convective heat transfer.

    This object needs to store:
      1) Fluid name
      2) A map between a ThermalMaterial name and the corresponding
         temperature-dependent film coefficient and its derivative
    """

    def get_dict(self):
        """
        Returns the data as a dictionary
        """
        raise NotImplementedError()

    def get_type(self):
        """
        Return the string type for the data
        """
        raise NotImplementedError()

    @classmethod
    def load(cls, fname, modelname):
        """
        Load a FluidMaterial object from a file

        Parameters:
          fname       file name to load from
        """
        tag, typ = find_name(fname, modelname)
        data = load_node(tag)[modelname]

        if typ == "PiecewiseLinearFluidMaterial":
            return PiecewiseLinearFluidMaterial.load(data)
        elif typ == "ConstantFluidMaterial":
            return ConstantFluidMaterial.load(data)
        else:
            raise ValueError("Unknown FluidMaterial type %s" % typ)

    def save(self, fname, modelname):
        """
        Save the model to an XML file as modelname

        Parameters:
          fname       filename to use
          modelname   (base tag) to use
        """
        root = ET.Element("models")

        save_node(modelname, self.get_dict(), root, attrib={"type": self.get_type()})

        tree = ET.ElementTree(element=root)
        tree.write(fname)


class ConstantFluidMaterial(FluidMaterial):
    """
    Supply a mapping between the material type and a constant
    film coefficient
    """

    def __init__(self, data):
        """
        Dictionary of the form {name: value} mapping
        a material name to the definition of the piecewise linear map

        Parameters:
          data:       the dictionary
        """
        self.data = data

    def get_dict(self):
        return {k: str(v) for k, v in self.data.items()}

    def get_type(self):
        return "ConstantFluidMaterial"

    @classmethod
    def load(cls, values):
        """
        Load from a dictionary

        Parameters:
          values      dictionary data
        """
        data = {k: float(val) for k, val in values.items()}
        return cls(data)

    def coefficient(self, material, T):
        """
        Return the film coefficient for the given material and temperature

        Parameters:
          material:       material name
          T:              temperatures

        """
        if material in self.data:
            return T * 0.0 + self.data[material]
        else:
            return T * 0.0 + self.data["default"]

    # pylint: disable=unused-argument
    def dcoefficient(self, material, T):
        """
        Return the derivative of the film coefficient with respect to
        temperature for the give material and temperature.

        Parameters:
          material:       material name
          T:              temperatures
        """
        return T * 0.0


class PiecewiseLinearFluidMaterial(FluidMaterial):
    """
    Supply a mapping between the material type and a piecewise linear
    interpolate defining the film coefficient as a function of temperature.
    """

    def __init__(self, data):
        """
        Dictionary of the form {name: (temperatures, values)} mapping
        a material name to the definition of the piecewise linear map

        Parameters:
          data:       the dictionary
        """
        self.data = data

        self.fns = {name: make_piecewise(T, v) for name, (T, v) in data.items()}

    def get_dict(self):
        return {
            k: {"temp": string_array(T), "values": string_array(v)}
            for k, (T, v) in self.data.items()
        }

    def get_type(self):
        return "PiecewiseLinearFluidMaterial"

    @classmethod
    def load(cls, values):
        """
        Load from a dictionary

        Parameters:
          values      dictionary data
        """
        data = {
            k: (destring_array(pair["temp"]), destring_array(pair["values"]))
            for k, pair in values.items()
        }
        return cls(data)

    def coefficient(self, material, T):
        """
        Return the film coefficient for the given material and temperature

        Parameters:
          material:       material name
          T:              temperatures

        """
        if material in self.fns.keys():
            return self.fns[material][0](T)
        else:
            return self.fns["default"][0](T)

    def dcoefficient(self, material, T):
        """
        Return the derivative of the film coefficient with respect to
        temperature for the give material and temperature.

        Parameters:
          material:       material name
          T:              temperatures
        """
        if material in self.fns.keys():
            return self.fns[material][1](T)
        else:
            return self.fns["default"][1](T)


class StructuralMaterial:
    """
    Properties for structural metallic material

    Supply
      1) cycles to failure as a function of temperature and strain range
      2) time to rupture as a function of temperaure and stress
      3) checks creep-fatigue interaction diagram
    """

    def __init__(self, data):
        self.data = data

    def cycles_to_fail(self, pname, temp, erange):
        """
        Returns fatigue cycles to failure at a given temperature and strain range

        Parameters:
          pname:       property name ("nominalFatigue")
          erange:      strain range in mm/mm
          temp:        temperature in K
        """
        pdata = self.data[pname]
        T, a, n, cutoff = [], [], [], []

        for i in pdata:
            T.append(destring_array(pdata[i]["T"]))
            a.append(destring_array(pdata[i]["a"]))
            n.append(destring_array(pdata[i]["n"]))
            cutoff.append(destring_array(pdata[i]["cutoff"]))

            if np.array(a).shape != np.array(n).shape:
                raise ValueError("\tThe lists of a and n must have equal lengths!")

        inds = np.array(T).argsort(axis=0)
        T = np.array(T)[inds]
        a = np.array(a)[inds]
        n = np.array(n)[inds]
        cutoff = np.array(cutoff)[inds]

        if temp > max(T):
            raise ValueError(
                "\ttemperature is out of range for cycle to failure determination"
            )

        for i in range(np.size(T, axis=0)):
            if temp <= T[i]:
                polysum = 0.0
                if erange <= cutoff[i]:
                    erange = cutoff[i][0][0]
                for (b, m) in zip(a[i][0], n[i][0]):
                    polysum += b * np.log10(erange) ** m
                break

        return 10**polysum

    def time_to_rupture(self, pname, temp, stress):
        """
        Returns time to rupture at a given temperature and stress

        Parameters:
          pname:       property name ("averageRupture" or "lowerboundRupture")
          stress:      stress in MPa
          temp:        temperature in K
        """
        pdata = self.data[pname]

        a = destring_array(pdata["a"])
        n = destring_array(pdata["n"])
        C = destring_array(pdata["C"])

        if a.shape != n.shape:
            raise ValueError("The lists of a and n must have equal lengths!")

        if stress.shape != temp.shape:
            raise ValueError("Stress and temperature must have the same shape!")

        zeros = stress == 0.0
        not_zeros = np.logical_not(zeros)

        res = np.zeros(stress.shape)
        for (b, m) in zip(a, n):
            res[not_zeros] += b * np.log10(stress[not_zeros]) ** m
        res[not_zeros] = 10.0 ** (res[not_zeros] / temp[not_zeros] - C)
        res[zeros] = np.inf

        return res

    def inside_envelope(self, pname, damage_fatigue, damage_creep):
        """
        Returns True if the point lies inside the design envelope and False if not

        Parameters:
          pname:               property name ("cfinteraction")
          damage_fatigue:      fatigue damage fraction
          creep_fatigue:       creep damage fraction
        """

        if damage_fatigue < 0.0 or damage_creep < 0.0:
            raise ValueError("\tout of range: negative damage fraction")

        pdata = destring_array(self.data[pname])

        x_1 = 0.0
        y_1 = 1.0
        x_2 = pdata[0]
        y_2 = pdata[1]
        x_3 = 1.0
        y_3 = 0.0

        if damage_fatigue < x_2:
            return damage_creep <= (
                (y_2 - y_1) / (x_2 - x_1) * (damage_fatigue - x_1) + y_1
            )
        return damage_creep <= (
            (y_3 - y_2) / (x_3 - x_2) * (damage_fatigue - x_2) + y_2
        )

    @classmethod
    def load(cls, fname, model):
        """
        Load a Structural Material object from a file

        Parameters:
          fname:       file name to load from
          material:    model name
        """
        tag = ET.parse(fname).getroot().find(model)
        return cls(load_node(tag)[model])

    def save(self, fname, modelname):
        """
        Save to a particular file under a particular model name
        """
        root = ET.Element("models")
        save_node(modelname, self.data, root)
        tree = ET.ElementTree(element=root)
        tree.write(fname)


def make_piecewise(x, y):
    """
    Make two piecewise interpolation functions: a piecewise linear
    interpolate between x and y and the corresponding derivative.
    """
    ydiff = np.zeros(y.shape)
    ydiff[:-1] = np.diff(y) / np.diff(x)
    ydiff[-1] = ydiff[-2]

    return inter.interp1d(x, y, fill_value=(ydiff[0], ydiff[-1])), inter.interp1d(
        x, ydiff, kind="previous", fill_value=(0, 0)
    )


def find_name(xmlfile, name):
    """
    Find the base tag with name in an XML file

    Parameters:
      xmlfile:        file name
      name:           tag to look for
    """
    root = ET.parse(xmlfile).getroot()
    tag = root.find(name)

    return tag, tag.attrib["type"]


def save_node(name, entry, node, attrib={}):
    """
    Save a dictionary to a particular node

    Parameters:
      name:     name of the new node
      entry:    entry of interest
      node:     ET parent node object

    Additional parameters:
      attribs   node attributes
    """
    nnode = ET.SubElement(node, name, attrib)
    if isinstance(entry, dict):
        for k, v in entry.items():
            save_node(k, v, nnode)
    else:
        nnode.text = entry


def load_node(node):
    """
    The actual function that does the loading by walking the XML file

    Parameters:
      node:      xml node object from ET
    """
    if len(node) > 0:
        return {node.tag: dict(ChainMap(*(load_node(child) for child in node)))}
    else:
        return {node.tag: node.text}


def string_array(array):
    """
    Make a numpy array a space separated string
    """
    return " ".join(map(str, array))


def destring_array(string):
    """
    Make an array from a space separated string
    """
    return np.array(list(map(float, string.split(" "))))


class CeramicMaterial:
    """
    Properties for structural ceramic material

    Supply
      1) Weibull strength as a function of temperature
      2) Weibull modulus as a function of temperature
      3) c_bar mode mixity parameter
      ... expand later for time-dependent properties
    """

    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def load(cls, fname, model):
        """
        Load a Ceramic material from a file

        Parameters:
          fname:       file name to load from
          material:    model name
        """
        tag, typ = find_name(fname, model)

        if typ == "StandardModel":
            return StandardCeramicMaterial.load(tag)
        else:
            raise ValueError("Unknown ceramic model type %s in damage data!" % typ)


class StandardCeramicMaterial:
    """
    Ceramic material where:

    1) Weibull strength depends on temperature
    2) Weibull modulus depends on temperature
    3) Constant c_bar parameter
    4) Constant Poisson's ratio
    """

    def __init__(
        self,
        s_temperatures,
        strengths,
        m_temperatures,
        modulus,
        c_bar,
        nu,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.s0 = inter.interp1d(s_temperatures, strengths)
        self.s_temperatures = s_temperatures
        self.strengths = strengths
        self.m_temperatures = m_temperatures
        self.mvals = modulus
        self.m = inter.interp1d(m_temperatures, modulus)
        self.C = c_bar
        self.nu_val = nu

    def strength(self, T):
        """
        Weibull strength as a function of temperature
        """
        return self.s0(T)

    def modulus(self, T):
        """
        Weibull modulus as a function of temperature
        """
        return self.m(T)

    def c_bar(self, T):
        """
        Mode mixity parameter as a function of temperature
        """
        if np.isscalar(T):
            return self.C
        else:
            return self.C * np.ones(T.shape)

    def nu(self, T):
        """
        Poisson's ratio as a function of temperature
        """
        if np.isscalar(T):
            return self.nu_val
        else:
            return self.nu_val * np.ones(T.shape)

    @classmethod
    def load(cls, node):
        """
        Load a Ceramic material from a file

        Parameters:
          node:    node with model
        """
        strength = node.find("strength")
        c_bar = node.find("c_bar")
        s_temps = strength.find("temperatures")
        svals = strength.find("values")
        m = node.find("modulus")
        m_temps = m.find("temperatures")
        mvals = m.find("values")
        nu = node.find("nu")

        return StandardCeramicMaterial(
            np.array(list(map(float, s_temps.text.strip().split()))),
            np.array(list(map(float, svals.text.strip().split()))),
            np.array(list(map(float, m_temps.text.strip().split()))),
            np.array(list(map(float, mvals.text.strip().split()))),
            float(c_bar.text),
            float(nu.text),
        )

    def save(self, fname, modelname):
        """
        Save to a particular file under a particular model name
        """
        root = ET.Element("models")

        base = ET.SubElement(root, modelname, {"type": "StandardModel"})
        strength = ET.SubElement(base, "strength")
        temps = ET.SubElement(strength, "temperatures")
        temps.text = " ".join(map(str, self.s_temperatures))
        svals = ET.SubElement(strength, "values")
        svals.text = " ".join(map(str, self.strengths))

        m = ET.SubElement(base, "modulus")
        mtemps = ET.SubElement(m, "temperatures")
        mtemps.text = " ".join(map(str, self.m_temperatures))
        mvals = ET.SubElement(m, "values")
        mvals.text = " ".join(map(str, self.mvals))

        c_bar = ET.SubElement(base, "c_bar")
        c_bar.text = str(self.C)

        nu = ET.SubElement(base, "nu")
        nu.text = str(self.nu_val)

        tree = ET.ElementTree(element=root)
        tree.write(fname)



"""
  Module that facilitates loading in all the material data
"""
import os.path

import xml.etree.ElementTree as ET

from srlife import materials
from srlife.thermohydraulics import thermalfluid

LIBRARY_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


def get_file(directory, name):
    """Return a file path or raise an error

    Args:
      directory (str): directory the file should be in
      name (str): actual file name

    Raises:
      RunetimeError: if the file does not exist
    """
    attempt = os.path.join(directory, name + ".xml")
    if not os.path.exists(attempt):
        raise RuntimeError("Material with name %s does not exists in database!" % name)
    return attempt


def load_fluid(name, model):
    """Load fluid material properties:

    Args:
      name (str): name of the fluid (title of xml file)
      model (str):  particular convection model to use

    Returns:
      material.FluidMaterial: fluid material model
    """
    fdir = os.path.join(LIBRARY_DIR, "fluid")
    filename = get_file(fdir, name)
    return materials.FluidMaterial.load(filename, model)


def load_thermal_fluid(name, model):
    """Load thermal fluid material properties

    Args:
        name (str): name of the fluid (XML filename)
        model (str): model variant to use
    """
    fdir = os.path.join(LIBRARY_DIR, "thermalfluid")
    filename = get_file(fdir, name)
    return thermalfluid.ThermalFluidMaterial.load(filename, model)


def load_material(name, thermal_model, deformation_model, damage_model):
    """Load solid material properties

    Args:
      name (str): name of the material (title of xml file)
      thermal_model (str): which thermal model variant to use
      deformation_model (str): which deformation model variant to use
      damage_model (str): which damage model variant to use

    Returns:
      materials.ThermalMaterial: thermal material model
      materials.DeformationMaterial: deformation material model
      materials.StructuralMaterial: damage material model
    """
    return (
        load_thermal(name, thermal_model),
        load_deformation(name, deformation_model),
        load_damage(name, damage_model),
    )


def load_thermal(name, model):
    """Load thermal material data

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.ThermalMaterial: thermal material model
    """
    fdir = os.path.join(LIBRARY_DIR, "thermal")
    filename = get_file(fdir, name)
    return materials.ThermalMaterial.load(filename, model)


def load_deformation(name, model):
    """Load a deformation model from file

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.DeformationMaterial: deformation material model
    """
    fdir = os.path.join(LIBRARY_DIR, "deformation")
    filename = get_file(fdir, name)
    return materials.DeformationMaterial(filename, model)


def load_damage(name, model):
    """Load a damage model from file

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.StructuralMaterial: damage material model
    """
    fdir = os.path.join(LIBRARY_DIR, "damage")
    filename = get_file(fdir, name)

    mat_type = get_type(filename)

    if mat_type == "metallic":
        return materials.StructuralMaterial.load(filename, model)
    elif mat_type == "ceramic":
        return materials.CeramicMaterial.load(filename, model)
    else:
        raise ValueError("Unknown material type %s in XML damage file." % mat_type)


def get_type(filename):
    """
    Report if this is a metallic or ceramic material
    """
    return ET.parse(filename).getroot().attrib["type"]



# pylint: disable=no-member

"""
  Meta-structural solver defining a network of nonlinear springs
"""

from abc import ABC, abstractmethod

import numpy as np
import networkx as nx
import multiprocess

from srlife.solvers import newton


class Spring(ABC):
    """
    Spring base class defining interface.

    Class will accommodate full nonlinear tubes (through FEA) as well
    as simple linear springs
    """

    @abstractmethod
    def force_and_stiffness(self, i, d):
        """
        Return the current forces and stiffness for some displacement

        Parameters:
          i           timestep
          d           displacement
        """
        return

    @abstractmethod
    def update_state(self, i):
        """
        Update to the next state after a successful solve

        Parameters:
          i           universal timestep number
        """
        return


class LinearSpring(Spring):
    """
    A linear elastic spring
    """

    def __init__(self, k):
        """
        Parameters:
          k       spring stiffness
        """
        self.k = k
        self.state_np1 = None
        self.state_n = None

    def force_and_stiffness(self, i, d):
        """
        Simple linear update

        Parameters:
          i       universal timestep index
          d       current displacement
        """
        return self.k * d, self.k

    def update_state(self, i):
        """
        We don't have state!

        Parameters:
          i       universal timestep number
        """
        return


class TubeSpring(Spring):
    """
    A spring object encapsulating the entire tube thermo-structural solver
    """

    def __init__(self, tube, solver, material):
        """
        Parameters:
          tube        tube object, with thermal history defined
          solver      structural solver to use
          material    tube material model (NEML)
        """
        self.tube = tube
        self.solver = solver
        self.material = material

        # Setup the tube to receive results
        self.solver.setup_tube(self.tube)

        # Get the first state
        self.state_n = self.solver.init_state(self.tube, self.material, i=0)

        # Dump this first state to the Tube
        self.solver.dump_state(self.tube, 0, self.state_n)

    def force_and_stiffness(self, i, d):
        """
        Parameters:
          d       current displacement
        """
        self.state_np1 = self.solver.solve(self.tube, i, self.state_n, d)

        return self.state_np1.force, self.state_np1.stiffness

    def update_state(self, i):
        """
        Parameters:
          i       timestep
        """
        self.solver.dump_state(self.tube, i, self.state_np1)
        self.state_n = self.state_np1


class SpringNetwork(nx.MultiGraph):
    """
    This mimics the NEML class with some additional features.

    Graph structure:
      Nodes are displacements
      Edges can be:
        Bars
        "disconnect"
        "rigid"

      Where "free" and "fixed" get factored out in the final network

    Boundary conditions:
      Live at nodes
      Can be imposed forces or displacements (though only displacements are
        going to bused in the end)
      Are simply tuples of (type, function) where
        type:
          force
          displacement
        function:
          function of time giving the values

    The object takes the time values from individual tubes by default, but
    this can be overrode with new step values.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.times = None
        self.displacements = None
        self.atol = kwargs.pop("atol", 1.0e-8)
        self.rtol = kwargs.pop("rtol", 1.0e-6)
        self.miter = kwargs.pop("miter", 25)
        self.verbose = kwargs.pop("verbose", False)

    def __copy(self, other):
        """
        Copy ancillary information from one network to another

        Parameters:
          other       other network
        """
        other.times = self.times
        other.rtol = self.rtol
        other.atol = self.atol
        other.miter = self.miter
        other.verbose = self.verbose

    def set_times(self, times):
        """
        Set the discrete times to a list

        Parameters:
          times       new times
        """
        self.times = times

    def validate_setup(self):
        """
        Validate the data structure
        """
        if self.times is None:
            self.__set_default_times()

        # Validate the type of each edge and the internal times (if provided)
        for _, _, edge in self.edges(data=True):
            if isinstance(edge["object"], TubeSpring):
                if not np.allclose(edge["object"].tube.times, self.times):
                    raise ValueError(
                        "A tube object has discrete times that are"
                        " not consistent with the spring network!"
                    )

            if not isinstance(edge["object"], (Spring, str)):
                raise ValueError("An edge object is not either a spring or a string!")

            if isinstance(edge["object"], str) and edge["object"] not in (
                "disconnect",
                "rigid",
            ):
                raise ValueError("Special edge type %s not recognized!" % edge)

    def __set_default_times(self):
        """
        Grab times from the first Tube object or raise error
        """
        for _, _, edge in self.edges(data=True):
            if isinstance(edge["object"], TubeSpring):
                self.times = edge["object"].tube.times
                break
        else:
            raise ValueError("No times provided and no Tube objects to copy from!")

    def force_bc(self, node, function):
        """
        Add a force boundary condition

        Parameters:
          node        location
          function    force as a function of time
        """
        self.nodes[node]["bc"] = ("force", function)

    def displacement_bc(self, node, function):
        """
        Add a displacement boundary condition

        Parameters:
          node        location
          function    displacement as a function of time
        """
        self.nodes[node]["bc"] = ("displacement", function)

    def reduce_graph(self):
        """
        Reduce the graph structure to a list of graphs which can be
        solved by:

        1) Removing rigid links
        2) Splitting the graph at disconnect
        """
        self.remove_rigid()
        return self.split_disconnect()

    def remove_rigid(self):
        """
        Remove rigid links by merging the nodes in place
        """
        while True:
            for i, j, key, edge in self.edges(data=True, keys=True):
                # Can't handle duplicate springs
                if edge["object"] == "rigid":
                    if self.number_of_edges(i, j) != 1:
                        raise RuntimeError("Cannot have a rigid link across a spring!")
                    bci = "bc" in self.nodes[i]
                    bcj = "bc" in self.nodes[j]
                    if bci and bcj:
                        raise RuntimeError("Cannot merge two nodes with BCs!")
                    self.remove_edge(i, j, key=key)

                    if i < j:
                        if bcj:
                            raise RuntimeError("Internal error: deleting BC")
                        self.replace(nx.contracted_nodes(self, i, j))
                    else:
                        if bci:
                            raise RuntimeError("Internal error: deleting BC")
                        self.replace(nx.contracted_nodes(self, j, i))
                    break
            else:
                break

    def replace(self, other):
        """
        Replace the graph data structure with a new graph
        """
        self.clear()
        self.update(other)

    def split_disconnect(self):
        """
        Split the graph into disjoint parts by taking apart at the disconnects
        """
        while True:
            for i, j, key, edge in self.edges(data=True, keys=True):
                if edge["object"] == "disconnect":
                    self.remove_edge(i, j, key=key)
                    break
            else:
                break

        components = list(nx.connected_components(self))

        nobjs = [SpringNetwork(self.subgraph(nset)) for nset in components]
        for obj in nobjs:
            self.__copy(obj)

        fobjs = []
        for obj in nobjs:
            if obj.size() != 0:
                fobjs.append(obj)

        return fobjs

    def validate_solve(self):
        """
        Validate that the graph is ready to actually solve

        Conditions:
          1) All dummy edges removed
          2) At least one fixed condition
          3) All nodes sequentially numbers from zero
        """
        # Check that we have times
        if self.times is None:
            raise RuntimeError("Discrete times must be established at solve!")

        # Check that everything is a spring
        for _, _, edge in self.edges(data=True):
            if not isinstance(edge["object"], Spring):
                raise RuntimeError("All edges must be springs at solve!")

        # Check that there is at least one fixed boundary
        ndisp = 0
        for node in self.nodes():
            if "bc" in self.nodes[node]:
                if self.nodes[node]["bc"][0] == "displacement":
                    ndisp += 1

        if ndisp == 0:
            raise RuntimeError("Spring network requires at least one fixed BC!")

        # To to see if the graph is fully-connected
        if not nx.is_connected(self):
            raise RuntimeError("Spring network must be fully connected at solve!")

    def solve(self, i, nthreads=1):
        """
        Solve discrete step i

        Parameter:
          i           discrete time step

        Additional parameters:
          nthreads    number of threads to use
        """
        # Keep forces and displacements for debug
        if self.displacements is None:
            self.displacements = np.zeros((len(self.nodes),))

        # Make sure we can solve
        self.validate_solve()

        # Set the time step
        self.i = i

        # Figure out free/fixed displacements
        (
            self.dmap,
            self.free,
            self.forces,
            self.fixed,
            self.fixed_displacements,
        ) = self.dof_maps(i)

        # Actually solve
        d = newton(
            lambda x: self.RJ(x, nthreads=nthreads),
            self.displacements[self.dmap[self.free]],
            verbose=self.verbose,
            rel_tol=self.rtol,
            abs_tol=self.atol,
            miters=self.miter,
        )

        # Store the displacements, for fun
        self.displacements[self.dmap[self.free]] = d
        self.displacements[self.dmap[self.fixed]] = self.fixed_displacements

        # Advance the state
        for _, _, edge in self.edges(data=True):
            edge["object"].update_state(self.i)

    def RJ(self, d, nthreads=1):
        """
        Actually calculate the residual and Jacobian equation for the step

        Parameters:
          d       free displacements
        """
        dall = np.zeros((len(self.nodes),))
        dall[self.dmap[self.fixed]] = self.fixed_displacements
        dall[self.dmap[self.free]] = d

        if nthreads > 1:
            with multiprocess.Pool(nthreads) as p:
                res = list(p.map(lambda e: self.fj(dall, *e), self.edges(data=True)))
        else:
            res = list(map(lambda e: self.fj(dall, *e), self.edges(data=True)))

        Fint = sum(r[0] for r in res)
        J = sum(r[1] for r in res)

        for k, (_, _, edge) in enumerate(self.edges(data=True)):
            edge["object"].state_np1 = res[k][2]

        return (
            Fint[self.dmap[self.free]] - self.forces,
            J[self.dmap[self.free], :][:, self.dmap[self.free]],
        )

    def fj(self, dall, i, j, edge):
        """
        Calculate the force and jacobian contributions for a single edge
        """
        ii = self.dmap[i]
        jj = self.dmap[j]
        ss = np.sign(ii - jj)
        Fint = np.zeros((len(self.nodes),))
        J = np.zeros((len(self.nodes), len(self.nodes)))
        f, k = edge["object"].force_and_stiffness(self.i, (dall[jj] - dall[ii]) * ss)
        Fint[ii] += -f
        Fint[jj] += f

        J[ii, ii] += k
        J[ii, jj] += -k
        J[jj, ii] += -k
        J[jj, jj] += k

        return Fint * ss, J, edge["object"].state_np1

    def dof_maps(self, i):
        """
        Get the degree of freedom maps for the given time step

        Parameters:
          i       discrete time step
        """
        free = []
        fixed = []
        forces = []
        displacements = []
        time = self.times[i]
        nodes = []
        for node in sorted(self.nodes):
            nodes.append(node)
            if "bc" in self.nodes[node]:
                if self.nodes[node]["bc"][0] == "displacement":
                    fixed.append(node)
                    displacements.append(self.nodes[node]["bc"][1](time))
                elif self.nodes[node]["bc"][0] == "force":
                    free.append(node)
                    forces.append(self.nodes[node]["bc"][1](time))
                else:
                    raise ValueError("Unknown BC type %s!" % self.nodes[node]["bc"][0])
            else:
                free.append(node)
                forces.append(0)

        nodes = np.array(nodes, dtype=int)
        total = (np.array(range(0, max(nodes) + 1), dtype=int) * 0) - 1
        rnums = np.array(range(len(nodes)), dtype=int)
        total[nodes] = rnums

        return (
            total,
            np.array(free, dtype=int),
            np.array(forces),
            np.array(fixed, dtype=int),
            np.array(displacements),
        )

    def solve_all(self, nthreads=1, decorator=lambda x, nitems: x):
        """
        Solve all time steps

        Additional parameters:
          nthreads:       number of threads to use
          decorator:      progress bar decorator
        """
        self.validate_solve()
        list(
            decorator(
                map(lambda i: self.solve(i, nthreads), range(1, len(self.times))),
                len(self.times) - 1,
            )
        )
        return self



import unittest
import tempfile
import os, os.path

import numpy as np

from srlife import receiver, thermal, materials


def make_tube(D, period):
    R = 1.0
    t = 0.1
    h = 10.0
    nr = 10
    nt = 20
    nz = 10
    T0 = 50

    tmax = period

    ntime = 10

    T0 = 300

    tube = receiver.Tube(R, t, h, nr, nt, nz, T0=T0)
    if D == 1:
        tube.make_1D(h / 2, 1)
    elif D == 2:
        tube.make_2D(h / 2)

    times = np.linspace(0, tmax, ntime)

    tube.set_times(times)

    Tf_0 = 300
    Tf_m = 600

    def fluid_T(t):
        if t < tmax / 2.0:
            return Tf_0 + (Tf_m - Tf_0) / (tmax / 2.0) * t
        else:
            return Tf_m - (Tf_m - Tf_0) / (tmax / 2.0) * (t - tmax / 2.0)

    ftemps = np.array([fluid_T(t) * np.ones((nz,)) for t in times])

    inner_convection = receiver.ConvectiveBC(R - t, h, nz, times, ftemps)
    tube.set_bc(inner_convection, "inner")

    hflux = receiver.HeatFluxBC(R, h, nt, nz, times, np.zeros((ntime, nt, nz)) + 2.0)
    tube.set_bc(hflux, "outer")

    solver = thermal.FiniteDifferenceImplicitThermalSolver()

    tmat = materials.ConstantThermalMaterial("dummy", 20.0e-3, 4.8)
    fmat = materials.ConstantFluidMaterial({"dummy": 8.1e-3})

    solver.solve(tube, tmat, fmat)

    return tube


def make_panel(D, ntube, period, stiff):
    panel = receiver.Panel(stiff)

    # To save some time
    t = make_tube(D, period)

    for i in range(ntube):
        panel.add_tube(t)

    return panel


def make_receiver(
    D,
    npanel,
    ntube,
    period=8 * 3600.0,
    days=1,
    panel_stiffness=0.0,
    manifold_stiffness=0.0,
):
    r = receiver.Receiver(period, days, panel_stiffness)
    for i in range(npanel):
        r.add_panel(make_panel(D, ntube, period, manifold_stiffness))

    return r


class TestVTKWriter(unittest.TestCase):
    def setUp(self):
        self.period = 8 * 3600.0
        self.npanel = 2
        self.ntube = 3
        self.days = 1
        self.ntime = 10
        self.panel_stiffness = 100.0
        self.manifold_stiffness = 100.0

    def test_1D(self):
        r = make_receiver(
            1,
            self.npanel,
            self.ntube,
            period=self.period,
            days=self.days,
            panel_stiffness=self.panel_stiffness,
            manifold_stiffness=self.manifold_stiffness,
        )
        tdir = tempfile.mkdtemp()

        r.write_vtk(os.path.join(tdir, "test"))

        nfile = len(os.listdir(tdir))

        self.assertEqual(nfile, self.npanel * self.ntube * self.ntime)

    def test_2D(self):
        r = make_receiver(
            2,
            self.npanel,
            self.ntube,
            period=self.period,
            days=self.days,
            panel_stiffness=self.panel_stiffness,
            manifold_stiffness=self.manifold_stiffness,
        )
        tdir = tempfile.mkdtemp()

        r.write_vtk(os.path.join(tdir, "test"))

        nfile = len(os.listdir(tdir))

        self.assertEqual(nfile, self.npanel * self.ntube * self.ntime)

    def test_3D(self):
        r = make_receiver(
            2,
            self.npanel,
            self.ntube,
            period=self.period,
            days=self.days,
            panel_stiffness=self.panel_stiffness,
            manifold_stiffness=self.manifold_stiffness,
        )
        tdir = tempfile.mkdtemp()

        r.write_vtk(os.path.join(tdir, "test"))

        nfile = len(os.listdir(tdir))

        self.assertEqual(nfile, self.npanel * self.ntube * self.ntime)



# Weirdness with pylint and vtk (some dynamic import thing going on there)
# pylint: disable=no-member
"""
  Module for various ways to write receiver data structures to disk in a human-readable way
"""
import vtk

import numpy as np


class VTKWriter:
    """
    Write a single tube file to VTK
    """

    def __init__(self, tube, fname):
        """
        Parameters:
          tube        tube object
          fname       base filename to use
        """
        self.tube = tube
        self.fname = fname

    def make_vtk_object(self):
        """
        Actually make the vtk object without the results
        """
        R, T, Z = self.tube.mesh

        X = R * np.cos(T)
        Y = R * np.sin(T)

        grid = vtk.vtkUnstructuredGrid()
        points = vtk.vtkPoints()

        for x, y, z in zip(X.flatten(), Y.flatten(), Z.flatten()):
            points.InsertNextPoint(x, y, z)

        grid.SetPoints(points)

        self._set_grid(grid)

        return grid

    def write(self):
        """
        Actually write the tube object to a vtk file using a
        VTKUntructuredGrid
        """
        R, T, Z = self.tube.mesh

        X = R * np.cos(T)
        Y = R * np.sin(T)

        for i in range(self.tube.ntime):
            grid = vtk.vtkUnstructuredGrid()
            points = vtk.vtkPoints()

            for x, y, z in zip(X.flatten(), Y.flatten(), Z.flatten()):
                points.InsertNextPoint(x, y, z)

            grid.SetPoints(points)

            self._set_grid(grid)

            writer = vtk.vtkXMLUnstructuredGridWriter()
            writer.SetFileName(self.fname + ("-%i" % i) + ".vtu")
            writer.SetInputData(grid)

            self._dump_point_data(grid, i)
            self._dump_element_data(grid, i)

            writer.Write()

    def _dump_point_data(self, grid, i):
        """
        Dump all our point data to the grid

        Parameters:
          grid        vtk unstructured grid
          i           time step
        """
        for field, data in self.tube.results.items():
            pdata = vtk.vtkFloatArray()
            pdata.SetNumberOfComponents(1)
            pdata.SetName(field)
            for d in data[i].flatten():
                pdata.InsertNextValue(d)
            grid.GetPointData().AddArray(pdata)

    def _dump_element_data(self, grid, i):
        for field, data in self.tube.quadrature_results.items():
            if field.split("_")[0] == "ghost":
                continue
            pdata = vtk.vtkFloatArray()
            pdata.SetNumberOfComponents(1)
            pdata.SetName(field)
            avg = np.mean(data[i], axis=1)
            for d in avg:
                pdata.InsertNextValue(d)
            grid.GetCellData().AddArray(pdata)

    def _set_grid(self, grid):
        """
        Setup the VTK unstructured grid for the mesh

        Parameters:
          grid        vtk unstructured grid
        """
        if self.tube.ndim == 1:
            self._set_grid_1d(grid)
        elif self.tube.ndim == 2:
            self._set_grid_2d(grid)
        else:
            self._set_grid_3d(grid)

    def _set_grid_1d(self, grid):
        """
        Setup the VTK unstructured grid for a 1d mesh

        Parameters:
          grid        vtk unstructured grid
        """
        flat = lambda i: i
        for i in range(self.tube.nr - 1):
            cell = vtk.vtkLine()
            for nk, a in enumerate((0, 1)):
                cell.GetPointIds().SetId(nk, flat(i + a))
            grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())

    def _set_grid_2d(self, grid):
        """
        Setup the VTK unstructured grid for a 1d mesh

        Parameters:
          grid        vtk unstructured grid
        """
        flat = lambda i, j: i * self.tube.nt + (j % self.tube.nt)
        for i in range(self.tube.nr - 1):
            for j in range(self.tube.nt):
                cell = vtk.vtkQuad()
                for nk, (a, b) in enumerate(((0, 0), (0, 1), (1, 1), (1, 0))):
                    cell.GetPointIds().SetId(nk, flat(i + a, j + b))
                grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())

    def _set_grid_3d(self, grid):
        """
        Setup the VTK unstructured grid for a 1d mesh

        Parameters:
          grid        vtk unstructured grid
        """
        flat = (
            lambda i, j, k: i * self.tube.nt * self.tube.nz
            + (j % self.tube.nt) * self.tube.nz
            + k
        )
        for i in range(self.tube.nr - 1):
            for j in range(self.tube.nt):
                for k in range(self.tube.nz - 1):
                    cell = vtk.vtkHexahedron()
                    for nk, (a, b, c) in enumerate(
                        (
                            (0, 0, 0),
                            (1, 0, 0),
                            (1, 1, 0),
                            (0, 1, 0),
                            (0, 0, 1),
                            (1, 0, 1),
                            (1, 1, 1),
                            (0, 1, 1),
                        )
                    ):
                        cell.GetPointIds().SetId(nk, flat(i + a, j + b, k + c))
                    grid.InsertNextCell(cell.GetCellType(), cell.GetPointIds())



import unittest
from parameterized import parameterized

import numpy as np

from srlife import materials, thermal, receiver

import sys

sys.path.append("test/solvers/thermal")
from thermalsol import ManufacturedSolution


class TestThermalManufactured(unittest.TestCase):
    def setUp(self):
        self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
        self.material = materials.ConstantThermalMaterial("Test", 10.0, 5.0)
        self.fluid = materials.ConstantFluidMaterial({"Test": 7.5})

        self.tol = 1e-2
        self.atol = 1e-3

    def _check_case(self, case):
        res = case.solve(self.solver, self.material, self.fluid)
        if not case.assess_comparison(res, self.tol, self.atol):
            print(case.name)
            self.assertTrue(False)

    @parameterized.expand(
        [
            (
                ManufacturedSolution(
                    "1D: no spatial",
                    1,
                    lambda t, r: t,
                    lambda t, k, alpha, r: k / alpha * (r * 0.0 + 1),
                ),
            ),
            (
                ManufacturedSolution(
                    "1D: spatial and time",
                    1,
                    lambda t, r: np.sin(t) * np.log(r),
                    lambda t, k, alpha, r: k / alpha * np.log(r) * np.cos(t),
                ),
            ),
            (
                ManufacturedSolution(
                    "1D: just spatial",
                    1,
                    lambda t, r: np.sin(r),
                    lambda t, k, alpha, r: k * np.sin(r) - k / r * np.cos(r),
                ),
            ),
            (
                ManufacturedSolution(
                    "1D: steady",
                    1,
                    lambda t, r: np.sin(r),
                    lambda t, k, alpha, r: k * np.sin(r) - k / r * np.cos(r),
                    steady=True,
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: just time",
                    2,
                    lambda t, r, th: t,
                    lambda t, k, alpha, r, th: k / alpha * (r * 0.0 + 1),
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: just r",
                    2,
                    lambda t, r, th: np.sin(r),
                    lambda t, k, alpha, r, th: k * np.sin(r) - k / r * np.cos(r),
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: just theta",
                    2,
                    lambda t, r, th: np.cos(th),
                    lambda t, k, alpha, r, th: k * np.cos(th) / r**2.0,
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: theta and r",
                    2,
                    lambda t, r, th: np.cos(th) / r,
                    lambda t, k, alpha, r, th: -k * np.cos(th) / r**3.0,
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: steady",
                    2,
                    lambda t, r, th: np.cos(th) / r,
                    lambda t, k, alpha, r, th: -k * np.cos(th) / r**3.0,
                    steady=True,
                ),
            ),
            (
                ManufacturedSolution(
                    "2D: all three",
                    2,
                    lambda t, r, th: np.log(r) * np.sin(th) / (t + 1),
                    lambda t, k, alpha, r, th: k
                    * np.log(r)
                    * np.sin(th)
                    / ((t + 1) * r**2.0)
                    - k / alpha * np.log(r) * np.sin(th) / (t + 1) ** 2.0,
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: just t",
                    3,
                    lambda t, r, th, z: t,
                    lambda t, k, alpha, r, th, z: k / alpha * (r * 0.0 + 1),
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: just r",
                    3,
                    lambda t, r, th, z: np.sin(r),
                    lambda t, k, alpha, r, th, z: k * np.sin(r) - k / r * np.cos(r),
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: just theta",
                    3,
                    lambda t, r, th, z: np.cos(th),
                    lambda t, k, alpha, r, th, z: k * np.cos(th) / r**2.0,
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: just z",
                    3,
                    lambda t, r, th, z: np.sin(z),
                    lambda t, k, alpha, r, th, z: k * np.sin(z),
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: all spatial",
                    3,
                    lambda t, r, th, z: z**2.0 * np.cos(th) / r,
                    lambda t, k, alpha, r, th, z: -k
                    * np.cos(th)
                    / r
                    * ((z / r) ** 2.0 + 2),
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: steady",
                    3,
                    lambda t, r, th, z: z**2.0 * np.cos(th) / r,
                    lambda t, k, alpha, r, th, z: -k
                    * np.cos(th)
                    / r
                    * ((z / r) ** 2.0 + 2),
                    steady=True,
                ),
            ),
            (
                ManufacturedSolution(
                    "3D: everything",
                    3,
                    lambda t, r, th, z: np.log(r) * np.sin(th) * np.cos(z) / (t + 1.0),
                    lambda t, k, alpha, r, th, z: k
                    * np.log(r)
                    * np.sin(th)
                    * np.cos(z)
                    / (t + 1)
                    * (1.0 + 1 / r**2.0 - 1.0 / (alpha * (t + 1))),
                ),
            ),
        ]
    )
    def test_case(self, case):
        self._check_case(case)


class TestThermalBCs(unittest.TestCase):
    def setUp(self):
        self.k = 15.0
        self.a = 5.0
        self.hcoef = 1.0

        self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
        self.material = materials.ConstantThermalMaterial("Test", self.k, self.a)
        self.fluid = materials.ConstantFluidMaterial({"Test": self.hcoef})

        self.t = 0.2
        self.r = 2.0
        self.h = 1.0
        self.nr = 10

        self.nt = 200
        self.nz = 2

        self.tmax = 1000.0
        self.ntime = 10

        self.tube = receiver.Tube(self.r, self.t, self.h, self.nr, self.nt, self.nz, 0)

        self.times = np.linspace(0, self.tmax, self.ntime + 1)
        self.tube.set_times(self.times)

        self.tube.make_1D(self.tube.h / 2, 0)

        self.ttimes, self.thetas, self.zs = np.meshgrid(
            self.times,
            np.linspace(0, 2 * np.pi, self.nt + 1)[:-1],
            np.linspace(0, self.h, self.nz),
            indexing="ij",
        )

        self.rs = np.linspace(self.r - self.t, self.r, self.nr)

    def test_dirichlet(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)

        Tleft = receiver.FixedTempBC(
            self.r - self.t,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.zeros(self.ttimes.shape),
        )
        self.tube.set_bc(Tleft, "inner")

        Tright = receiver.FixedTempBC(
            self.r, self.h, self.nt, self.nz, self.times, np.zeros(self.ttimes.shape)
        )
        self.tube.set_bc(Tright, "outer")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0)

        # Correct solution: temperatures head towards zero
        T = self.tube.results["temperature"][-1]
        self.assertTrue(np.allclose(T, 0))

    def test_neumann_left(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)
        q = 10.0

        Tleft = receiver.HeatFluxBC(
            self.r - self.t,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.ones(self.ttimes.shape) * q,
        )
        self.tube.set_bc(Tleft, "inner")

        Tright = receiver.FixedTempBC(
            self.r, self.h, self.nt, self.nz, self.times, np.zeros(self.ttimes.shape)
        )
        self.tube.set_bc(Tright, "outer")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0)

        # Correct solution:
        Tright = -q * (self.r - self.t) / self.k * np.log(self.rs) + q * (
            self.r - self.t
        ) / self.k * np.log(self.r)

        T = self.tube.results["temperature"][-1]

        self.assertTrue(np.allclose(T, Tright, rtol=1e-2))

    def test_neumann_right(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)
        q = 10.0

        Tright = receiver.HeatFluxBC(
            self.r, self.h, self.nt, self.nz, self.times, np.ones(self.ttimes.shape) * q
        )
        self.tube.set_bc(Tright, "outer")

        Tleft = receiver.FixedTempBC(
            self.r - self.t,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.zeros(self.ttimes.shape),
        )
        self.tube.set_bc(Tleft, "inner")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0)

        # Correct solution:
        Tright = q * self.r / self.k * np.log(self.rs) - q * self.r / self.k * np.log(
            self.r - self.t
        )

        T = self.tube.results["temperature"][-1]

        self.assertTrue(np.allclose(T, Tright, rtol=1e-2))

    def test_convection_left(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)

        T_inner = 25

        Tin = receiver.ConvectiveBC(
            self.r - self.t,
            self.h,
            self.nz,
            self.times,
            np.ones((self.ntime + 1, self.nz)) * T_inner,
        )
        self.tube.set_bc(Tin, "inner")

        Tright = receiver.FixedTempBC(
            self.r, self.h, self.nt, self.nz, self.times, np.zeros(self.ttimes.shape)
        )
        self.tube.set_bc(Tright, "outer")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0)

        T = self.tube.results["temperature"][-1]

        ri = self.r - self.t

        C1 = (
            T_inner
            * self.hcoef
            / (self.hcoef * np.log(ri) - self.hcoef * np.log(self.r) - self.k / ri)
        )
        C2 = -C1 * np.log(self.r)

        Texact = C1 * np.log(self.rs) + C2

        self.assertTrue(np.allclose(T, Texact, rtol=1.0e-2))

    def test_convection_right(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)

        T_outer = 25

        Tout = receiver.ConvectiveBC(
            self.r,
            self.h,
            self.nz,
            self.times,
            np.ones((self.ntime + 1, self.nz)) * T_outer,
        )
        self.tube.set_bc(Tout, "outer")

        Tleft = receiver.FixedTempBC(
            self.r - self.t,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.zeros(self.ttimes.shape),
        )
        self.tube.set_bc(Tleft, "inner")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0)

        T = self.tube.results["temperature"][-1]

        ri = self.r - self.t

        C1 = (
            self.hcoef
            * T_outer
            / (self.k / self.r + self.hcoef * np.log(self.r) - self.hcoef * np.log(ri))
        )
        C2 = -C1 * np.log(ri)

        Texact = C1 * np.log(self.rs) + C2

        self.assertTrue(np.allclose(T, Texact, rtol=1.0e-2))

    def test_substep(self):
        T0 = lambda x: np.sin(2 * np.pi * (x - self.r - self.t) / self.t)

        T_outer = 250

        Tout = receiver.ConvectiveBC(
            self.r,
            self.h,
            self.nz,
            self.times,
            np.ones((self.ntime + 1, self.nz)) * T_outer,
        )
        self.tube.set_bc(Tout, "outer")

        Tleft = receiver.FixedTempBC(
            self.r - self.t,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.zeros(self.ttimes.shape),
        )
        self.tube.set_bc(Tleft, "inner")

        self.solver.solve(self.tube, self.material, self.fluid, T0=T0, substep=10)

        T = self.tube.results["temperature"][-1]

        ri = self.r - self.t

        C1 = (
            self.hcoef
            * T_outer
            / (self.k / self.r + self.hcoef * np.log(self.r) - self.hcoef * np.log(ri))
        )
        C2 = -C1 * np.log(ri)

        Texact = C1 * np.log(self.rs) + C2

        self.assertTrue(np.allclose(T, Texact, rtol=1.0e-2))


class TestFunction(unittest.TestCase):
    def setUp(self):
        self.k = 15.0
        self.a = 5.0
        self.hcoef = 1.0

        self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
        self.material = materials.ConstantThermalMaterial("Test", self.k, self.a)
        self.fluid = materials.ConstantFluidMaterial({"Test": self.hcoef})

        self.t = 0.2
        self.r = 2.0
        self.h = 1.0

        self.nr = 5

        self.nt = 10
        self.nz = 5

        self.tmax = 10.0
        self.ntime = 10

        self.tube = receiver.Tube(self.r, self.t, self.h, self.nr, self.nt, self.nz, 0)

        self.times = np.linspace(0, self.tmax, self.ntime + 1)
        self.tube.set_times(self.times)

        self.T_inner = 100

        Tin = receiver.ConvectiveBC(
            self.r - self.t,
            self.h,
            self.nz,
            self.times,
            np.ones((self.ntime + 1, self.nz)) * self.T_inner,
        )
        self.tube.set_bc(Tin, "inner")

        Tright = receiver.HeatFluxBC(
            self.r,
            self.h,
            self.nt,
            self.nz,
            self.times,
            np.zeros((self.ntime + 1, self.nt, self.nz)),
        )
        self.tube.set_bc(Tright, "outer")

    def test_1D(self):
        self.tube.make_1D(self.tube.h / 2, 0)
        self.solver.solve(self.tube, self.material, self.fluid)

    def test_2D(self):
        self.tube.make_2D(self.tube.h / 2)
        self.solver.solve(self.tube, self.material, self.fluid)

    def test_3D(self):
        self.solver.solve(self.tube, self.material, self.fluid)



"""
  Nonlinear solvers and other accessories
"""

import numpy as np
import numpy.linalg as la

# pylint: disable=too-many-branches
def newton(
    RJ,
    x0,
    rel_tol=1.0 - 6,
    abs_tol=1.0e-8,
    miters=20,
    linear_solver=la.solve,
    verbose=True,
    return_extra=False,
    linesearch=True,
    max_search=10,
):
    """Simple newton-raphson solver

    Args:
      RJ:                           function that gives the residual and
                                    jacobian values
      x0:                           initial guess
      rel_tol (Optional[1.0e-6]):   relative convergence tolerance
      abs_tol (Optional[1.0e-8]):   absolute convergence tolerance
      miters (Optional[20]):        maximum number of iterations
      linear_solver (Optional[numpy.linalg.solve]): function that solves
                                                    the linear system A x = b
      verbose (Optional[True]):     if true, print debug info,
      return_extra (Optional[False]):     if true also return the final residual vector and Jacobian
      linesearch (Optional[True]):  if true do backtracking linesearch
      max_search (Optional[10]):    max number of backtracking steps
    """
    x = np.copy(x0)

    R, J = RJ(x)

    nR = la.norm(R)
    nR0 = nR

    if verbose:
        if linesearch:
            print("Iter\tnR\t\tnR/nR0\t\talpha")
            print("%i\t%3.2e\t%3.2e" % (0, nR0, 1))
        else:
            print("Iter\tnR\t\tnR/nR0")
            print("%i\t%3.2e" % (0, nR0))

    for i in range(miters):
        if nR < abs_tol or nR / nR0 < rel_tol:
            break
        dx = linear_solver(J, R)
        if linesearch:
            alpha = 1.0
            nR_last = nR
            x_last = np.copy(x)
            for _ in range(max_search):
                x = x_last - alpha * dx
                R, J = RJ(x)
                nR = la.norm(R)
                if nR < nR_last:
                    break
                alpha /= 2.0
        else:
            x -= dx
            R, J = RJ(x)
            nR = la.norm(R)

        if verbose:
            if linesearch:
                print("%i\t%3.2e\t%3.2e\t%3.2e" % (i + 1, nR, nR / nR0, alpha))
            else:
                print("%i\t%3.2e\t%3.2e" % (i + 1, nR, nR / nR0))
    else:
        raise RuntimeError("Nonlinear solver did not converge!")

    if verbose:
        print("")

    if return_extra:
        return x, R, J
    else:
        return x



"""
  Solution managers actually walk a model through all the steps to solve
"""
import multiprocess
import numpy as np

import tqdm

from srlife import solverparams, thermal


def merge_dicts(idict):
    """
    Merge an iterator over dictionaries into a single dict
    """
    return {k: v for d in idict for k, v in d.items()}


class SolutionManager:
    """Solution manager

    High level solution manager walking through the thermal, structural,
    and damage calculations.

    Args:
      receiver (receiver.Receiver): receiver object to solve
      thermal_solver (thermal.ThermalSolver): how to solve the heat transport
      thermal_material (materials.ThermalMaterial): solid thermal properties
      fluid_material (materials.FluidMaterial): fluid thermal properties
      structural_solver (structural.TubeSolver):  how to solve the mechanics problem
      deformation_material (materials.DeformationMaterial):  how things deform with time
      damage_material (materials.StructuralMaterial):  how to calculate creep damage
      system_solver (system.SystemSolver):  how to tie tubes into the structural system
      damage_model (damage.DamageCalculator): how to calculate damage from the results
      pset (Optional[solverparams.ParameterSet]): optional set of solver parameters
    """

    def __init__(
        self,
        receiver,
        thermal_solver,
        thermal_material,
        fluid_material,
        structural_solver,
        deformation_material,
        damage_material,
        system_solver,
        damage_model,
        pset=solverparams.ParameterSet(),
    ):
        self.receiver = receiver
        self.thermal_solver = thermal_solver
        self.thermal_material = thermal_material
        self.fluid_material = fluid_material
        self.structural_solver = structural_solver
        self.deformation_material = deformation_material
        self.damage_material = damage_material
        self.system_solver = system_solver
        self.damage_model = damage_model

        self.pset = pset
        self.nthreads = pset.get_default("nthreads", 1)
        self.progress = pset.get_default("progress_bars", False)

        self.page = pset.get_default("page_results", False)
        self.receiver.set_paging(self.page)

        self.heuristics = []

    def add_heuristic(self, heuristic):
        """Add a heuristic to use during the solve

        Args:
          heuristics (list): new heuristic to add
        """
        self.heuristics.append(heuristic)

    @property
    def tubes(self):
        """Direct iterator over tubes

        Returns:
          iterator over tubes
        """
        return self.receiver.tubes

    @property
    def ntubes(self):
        """Pass through to get the number of tubes that need to be analyzed

        Returns:
          int:  total number of tubes in entire receiver
        """
        return len(list(self.tubes))

    def progress_decorator(self, base, ntotal):
        """Either wrap with a progress bar decorator or return a dummy

        Args:
          base (function): base function to wrap
          ntotal (int): total number in iterator, needed for wrapping iterators

        Returns:
          either function wrapped with decorator or base function
        """
        if self.progress:
            return tqdm.tqdm(base, total=ntotal)
        else:
            return base

    def solve_life(self):
        """User interface: solve everything and return receiver life

        The trigger for everything: solve the complete problem and report the
        best-estimate life.

        Returns:
          float:  Number of allowed daily cycles
        """
        self.solve_heat_transfer()
        self.solve_structural()

        return self.calculate_damage()

    def calculate_damage(self):
        """Calculate damage from the results

        Returns:
          float:    Number of allowed daily cycles
        """
        if self.progress:
            print("Calculating damage:")
        return self.damage_model.determine_life(
            self.receiver,
            self.damage_material,
            nthreads=self.nthreads,
            decorator=self.progress_decorator,
        )

    def solve_reliability(self):
        """User interface: solve everything and return receiver reliability

        The trigger for everything: solve the complete problem and report the
        best-estimate reliability.

        Returns:
          float:  Reliability between 0 and 1
        """
        self.solve_heat_transfer()
        self.solve_structural()

        return self.calculate_reliability()

    def calculate_reliability(self):
        """Calculate reliability from the results

        Returns:
          float:    Reliability between 0 and 1
        """
        if self.progress:
            print("Calculating reliability :")
        return self.damage_model.determine_reliability(
            self.receiver,
            self.damage_material,
            nthreads=self.nthreads,
            decorator=self.progress_decorator,
        )

    def solve_heat_transfer(self):
        """Solve heat transfer for the receiver

        Adds thermal results in each tube
        """
        if isinstance(self.thermal_solver, thermal.ThermohydraulicsThermalSolver):
            if self.progress:
                print("Running thermohydraulic analysis")
            self.thermal_solver.solve_receiver(
                self.receiver,
                self.thermal_material,
                self.fluid_material,
                decorator=self.progress_decorator,
                nthreads=self.nthreads,
                **merge_dicts(
                    h.args_for_thermohydraulic_solver(self.receiver)
                    for h in self.heuristics
                )
            )
        else:
            self.solve_heat_transfer_tube()

    def solve_heat_transfer_tube(self):
        """Solve the heat transfer problem for each tube

        Adds the thermal results to each receiver.Tube object
        """
        # pylint: disable=no-member
        if self.progress:
            print("Running thermal analysis:")
        with multiprocess.Pool(self.nthreads) as p:
            temps = list(
                self.progress_decorator(
                    p.imap(
                        lambda x: self.thermal_solver.solve(
                            x,
                            self.thermal_material,
                            self.fluid_material,
                            **merge_dicts(
                                h.args_for_thermal_solver(self.receiver, x)
                                for h in self.heuristics
                            )
                        ),
                        self.tubes,
                    ),
                    self.ntubes,
                )
            )
        for tube, temps in zip(self.tubes, temps):
            tube.add_results("temperature", temps)

    def solve_structural(self):
        """Solve the structural problem for the complete system

        Adds the structural results to each tube.
        """
        if self.progress:
            print("Running structural analysis:")
        self.system_solver.solve(
            self.receiver,
            self.deformation_material,
            self.structural_solver,
            nthreads=self.nthreads,
            decorator=self.progress_decorator,
            verbose=self.progress,
        )


class Heuristic:
    """Solution heuristic superclass

    Class that defines a heuristic to modify the basic 3D solve process in some
    way.

    To implement a specific heuristic override the appropriate pure virtual
    methods.
    """

    def args_for_tube_thermal_solver(self, receiver, tube):
        """Add tube solver args

        Args:
          receiver (receiver.receiver):       receiver object affected
          tube (receiver.tube):               tube object affected

        """
        return {}

    def args_for_thermohydraulic_solver(self, receiver):
        """Add tube solver args

        Args:
          receiver (receiver.receiver):       receiver object affected

        """
        return {}


class CycleResetHeuristic(Heuristic):
    """
    Reset the tube temperatures each cycle to the initial values
    """

    def args_for_tube_thermal_solver(self, receiver, tube):
        """Modify the tube solver before the solve

        Args:
          receiver (receiver.receiver):       receiver object affected
          tube (receiver.tube):               tube object affected
        """
        resetter = thermal.TemperatureResetter(
            lambda t: np.isclose(t % receiver.period, 0), tube.T0
        )

        return {"resetters": [resetter]}

    def args_for_thermohydraulic_solver(self, receiver):
        """Add tube solver args

        Args:
          receiver (receiver.receiver):       receiver object affected

        """
        resetter = thermal.ReceiverResetter(
            lambda t: np.isclose(t % receiver.period, 0)
        )

        return {"resetters": [resetter]}



import unittest

import numpy as np

from srlife import spring


class TestSimple(unittest.TestCase):
    """
    Simple test cases with elastic springs
    """

    def test_single(self):
        network = spring.SpringNetwork()
        network.add_node(0)
        network.add_node(1)
        network.add_edge(0, 1, object=spring.LinearSpring(100))

        network.set_times([0, 1])

        network.validate_setup()

        network.displacement_bc(0, lambda t: 0.0)
        network.force_bc(1, lambda t: 100.0 * t)

        network.solve_all()

        self.assertTrue(np.allclose(network.displacements, [0, 1]))

    def test_system(self):
        k1 = 100.0
        k2 = 200.0
        k3 = 150.0
        f = 100.0

        network = spring.SpringNetwork()
        network.add_node(0)
        network.add_node(1)
        network.add_node(2)
        network.add_edge(0, 1, object=spring.LinearSpring(k1))
        network.add_edge(1, 2, object=spring.LinearSpring(k2))
        network.add_edge(1, 2, object=spring.LinearSpring(k3))

        network.set_times([0, 1])

        network.validate_setup()

        network.displacement_bc(2, lambda t: 0.0)
        network.force_bc(0, lambda t: t * f)

        network.solve_all()

        kp = k2 + k3
        dxp = f / kp
        dx1 = f / k1

        exact = np.array([dx1 + dxp, dxp, 0])

        self.assertTrue(np.allclose(network.displacements, exact))

    def test_system_thrads(self):
        k1 = 100.0
        k2 = 200.0
        k3 = 150.0
        f = 100.0

        network = spring.SpringNetwork()
        network.add_node(0)
        network.add_node(1)
        network.add_node(2)
        network.add_edge(0, 1, object=spring.LinearSpring(k1))
        network.add_edge(1, 2, object=spring.LinearSpring(k2))
        network.add_edge(1, 2, object=spring.LinearSpring(k3))

        network.set_times([0, 1])

        network.validate_setup()

        network.displacement_bc(2, lambda t: 0.0)
        network.force_bc(0, lambda t: t * f)

        network.solve_all(nthreads=2)

        kp = k2 + k3
        dxp = f / kp
        dx1 = f / k1

        exact = np.array([dx1 + dxp, dxp, 0])

        self.assertTrue(np.allclose(network.displacements, exact))

    def test_remove_rigid(self):
        network = spring.SpringNetwork()
        k = 100.0

        for i in range(6):
            network.add_node(i)

        network.add_edge(0, 1, object="rigid")
        network.add_edge(1, 3, object=spring.LinearSpring(k))
        network.add_edge(1, 3, object=spring.LinearSpring(k))
        network.add_edge(2, 4, object="rigid")
        network.add_edge(3, 5, object=spring.LinearSpring(k))
        network.add_edge(4, 5, object=spring.LinearSpring(k))

        network.displacement_bc(0, lambda x: 0)

        network.set_times([0, 1])
        network.validate_setup()

        network.remove_rigid()
        network.validate_solve()

        self.assertEqual(len(network.nodes), 4)
        self.assertEqual(len(network.edges), 4)

    def test_disconnect(self):
        network = spring.SpringNetwork()
        k = 100.0

        for i in range(8):
            network.add_node(i)

        network.add_edge(0, 1, object=spring.LinearSpring(k))
        network.add_edge(0, 2, object=spring.LinearSpring(k))

        network.add_edge(1, 3, object=spring.LinearSpring(k))
        network.add_edge(1, 3, object=spring.LinearSpring(k))
        network.add_edge(2, 4, object="disconnect")

        network.add_edge(3, 5, object=spring.LinearSpring(k))
        network.add_edge(4, 6, object=spring.LinearSpring(k))
        network.add_edge(4, 7, object=spring.LinearSpring(k))

        network.displacement_bc(0, lambda x: 0)
        network.displacement_bc(4, lambda x: 0)

        network.set_times([0, 1])
        network.validate_setup()

        subs = network.reduce_graph()

        self.assertEqual(len(subs), 2)

        for sub in subs:
            sub.validate_solve()



#!/usr/bin/env python3

import os.path

import numpy as np
import numpy.linalg as la
import scipy.io as sio

import matplotlib.pyplot as plt
from neml import models, elasticity, parse

import sys

sys.path.append("../../..")
from srlife import receiver, structural


class TestCase:
    def __init__(
        self,
        name,
        T,
        analytic,
        ri=0.9,
        ro=1.0,
        h=10.0,
        alpha=1.0e-5,
        E=100000.0,
        nu=0.3,
        p=1.0,
    ):
        self.name = name
        self.Tfn = T
        self.afn = analytic

        self.ri = ri
        self.ro = ro
        self.h = h

        self.alpha = alpha
        self.E = E
        self.nu = nu

        self.p = p

    def T(self, r):
        return self.Tfn(r, self.ri, self.ro)

    def exact(self, r):
        return self.afn(r, self.p, self.ri, self.ro, self.E, self.nu, self.alpha)

    def make_mat(self):
        emodel = elasticity.IsotropicLinearElasticModel(
            self.E, "youngs", self.nu, "poissons"
        )
        return models.SmallStrainElasticity(emodel, alpha=self.alpha)

    def make_tube(self, dim, nr=15, nt=30, nz=5):
        tube = receiver.Tube(self.ro, self.ro - self.ri, self.h, nr, nt, nz)

        if dim == 1:
            tube.make_1D(self.h / 2, 0)
        elif dim == 2:
            tube.make_2D(self.h / 2)

        times = np.array([0, 1])
        tube.set_times(times)

        R, _, _ = tube.mesh
        Ts = np.zeros((2,) + R.shape[:dim])
        Ts[1] = self.T(R)

        tube.add_results("temperature", Ts)

        if self.p != 0:
            tube.set_pressure_bc(receiver.PressureBC(times, times * self.p))

        return tube

    def run_comparison(self, dim, solver, axial_strain=0, nr=10, nt=20, nz=10):
        mat = self.make_mat()
        tube = self.make_tube(dim, nr, nt, nz)

        solver.setup_tube(tube)
        state_n = solver.init_state(tube, mat)

        state_np1 = solver.solve(tube, 1, state_n, axial_strain)

        solver.dump_state(tube, 1, state_np1)

        return tube

    def get_comparison(self, tube):
        if tube.ndim == 3:
            z = tube.nz // 2
            x_avg = np.mean(tube.results["disp_x"])
            u = tube.results["disp_x"][1, :, 0, z] - 2 * x_avg
            r = tube.mesh[0][:, 0, z]
        elif tube.ndim == 2:
            # The displacements tend to drift, need to recenter
            x_avg = np.mean(tube.results["disp_x"])
            u = tube.results["disp_x"][1, :, 0] - 2 * x_avg
            r = tube.mesh[0][:, 0, 0]
        else:
            u = tube.results["disp_x"][1]
            r = tube.mesh[0][:, 0, 0]

        return u, r

    def plot_comparison(self, tube):
        u, r = self.get_comparison(tube)

        plt.figure()
        plt.plot(r, u, "k-")
        plt.plot(r, self.exact(r), "k--")
        plt.xlabel("Radial position")
        plt.ylabel("Radial displacement")
        plt.title(self.name + ": " + "%iD" % tube.ndim)
        plt.show()

    def evaluate_comparison(self, tube):
        u, r = self.get_comparison(tube)

        err = np.abs(u - self.exact(r))
        rel = err / np.abs(self.exact(r))

        return np.max(err), np.max(rel)


def exact1(r, p, ri, ro, E, v, a):
    A = p / (1.0 / ro**2.0 - 1.0 / ri**2.0)
    C = -A / (2 * ro**2.0)

    res = (1.0 + v) / E * (-A / r + 2.0 * (1 - 2.0 * v) * C * r)

    return res


cases = [
    TestCase("Inner pressure", lambda T, ri, ro: 0.0, exact1, p=100, ri=8, ro=10.0)
]


# Moose comparison stuff
# MOOSE test stuff
ramp = lambda x: np.piecewise(
    x, [x < 1, x >= 1], [lambda xx: xx, lambda xx: 1.0 + xx * 0.0]
)
unit_temperature = (
    lambda x, y, z: (np.sqrt(x**2.0 + y**2.0) - 9)
    / (10 - 9)
    * np.cos(np.arctan2(y, x))
    * (z / 10.0 + 1)
)
temperature = lambda t, x, y, z: np.array(
    [100 * ramp(ti) * unit_temperature(x, y, z) for ti in t]
)
pressure = lambda t: 1.0 * ramp(t)

times = np.array([0, 0.5, 1, 101, 201, 301, 401, 501, 601, 701, 801, 901, 1001])
mat = parse.parse_xml(
    os.path.join(os.path.dirname(__file__), "moose-verification", "model.xml"),
    "creeping",
)

ri = 9.0
ro = 10.0
h = 10.0
nr = 11
nt = 20
nz = 6

moose_base = os.path.join(os.path.dirname(__file__), "moose-verification")
moose_ver = [os.path.join(moose_base, f) for f in ["1d_out.e", "2d_out.e", "3d_out.e"]]


def run_reference_simulation(d, solver):
    tube = receiver.Tube(ro, ro - ri, h, nr, nt, nz)
    tube.set_times(times)

    if d == 1:
        tube.make_1D(0, 0)
    elif d == 2:
        tube.make_2D(0)

    R, T, Z = tube.mesh
    X = R * np.cos(T)
    Y = R * np.sin(T)

    Ts = temperature(times, X, Y, Z).reshape((len(times),) + tube.dim[: tube.ndim])

    tube.add_results("temperature", Ts)
    tube.set_pressure_bc(receiver.PressureBC(times, pressure(times)))

    solver.setup_tube(tube)

    state_n = solver.init_state(tube, mat)

    for i in range(1, len(tube.times)):
        state_np1 = solver.solve(tube, i, state_n, 0.0)
        solver.dump_state(tube, i, state_np1)
        state_n = state_np1

    return tube


def load_displacements_exodus(efilename, d):
    nf = sio.netcdf_file(efilename)
    cnames = ["coordx", "coordy", "coordz"]
    coords = np.array([np.copy(nf.variables[c][:]) for c in cnames[:d]])

    names = [b"".join(v).decode("utf-8") for v in nf.variables["name_nod_var"]]
    disps = np.array(
        [
            np.copy(nf.variables["vals_nod_var%i" % (names.index(nm) + 1)][:])
            for nm in ["disp_x", "disp_y", "disp_z"][:d]
        ]
    )

    return coords, disps


def compare_nodal_field(c1, r1, c2, r2, d, dec=6, gtol=1.0e-8):
    c1 = np.round(c1, decimals=dec)
    c2 = np.round(c2, decimals=dec)
    for i in range(d - 1, -1, -1):
        one_args = c1[i, :].argsort(kind="mergesort")
        c1 = c1[:, one_args]

        two_args = c2[i, :].argsort(kind="mergesort")
        c2 = c2[:, two_args]

        for j in range(d):
            for k in range(r1.shape[1]):
                r1[j, k] = r1[j, k, one_args]
                r2[j, k] = r2[j, k, two_args]

    # Sanity check
    if la.norm(c1 - c2) > gtol:
        raise ValueError("Point clouds aren't equivalent...")

    error = np.abs(r1 - r2)
    aerror = np.max(error)

    keep = np.abs(r1) > gtol
    rerror = np.max(error[keep] / np.abs(r1[keep]))

    return aerror, rerror


def do_complete_comparison(d, solver):
    tube = run_reference_simulation(d, solver)

    R, T, Z = tube.mesh
    X = R * np.cos(T)
    Y = R * np.sin(T)
    cs = [X, Y, Z]
    coords = np.array([c.flatten() for c in cs[:d]])
    dres = ["disp_x", "disp_y", "disp_z"]
    disps = np.array([tube.results[d].reshape((len(tube.times), -1)) for d in dres[:d]])

    mc, md = load_displacements_exodus(moose_ver[d - 1], d)

    return compare_nodal_field(mc, md, coords, disps, d)


if __name__ == "__main__":
    solver = structural.PythonTubeSolver(verbose=False)

    print("Analytical comparison")
    print("=====================")
    print("")
    for d in range(1, 4):
        for case in cases:
            tube = case.run_comparison(d, solver)

            a, r = case.evaluate_comparison(tube)
            print(case.name + ": " "%iD" % d)
            print("Max absolute error: %e" % a)
            print("Max relative error: %e" % r)
            print("")

            case.plot_comparison(tube)

    print("MOOSE comparison")
    print("================")
    print("")
    for d in range(1, 4):
        a, r = do_complete_comparison(d, solver)
        print("%iD" % d)
        print("Max absolute error: %e" % a)
        print("")



"""
  Helper functions for converting to and from tensor notation
"""

from __future__ import division

import itertools

import numpy as np

mandel = ((0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1))
mandel_mults = (1, 1, 1, np.sqrt(2), np.sqrt(2), np.sqrt(2))


def make_M2T():
    """
    Make the tensor that goes from rank 4 Mandel to rank 4 tensors
    """
    R = np.zeros((3, 3, 3, 3, 6, 6))
    for a in range(6):
        for b in range(6):
            ind_a = itertools.permutations(mandel[a], r=2)
            ind_b = itertools.permutations(mandel[b], r=2)
            ma = mandel_mults[a]
            mb = mandel_mults[b]
            indexes = tuple(ai + bi for ai, bi in itertools.product(ind_a, ind_b))
            for ind in indexes:
                R[ind + (a, b)] = 1.0 / (ma * mb)

    return R


M2T = make_M2T().reshape(81, 36)


def make_usym():
    """
    Make the tensor that goes from Mandel to tensors
    """
    R = np.zeros((3, 3, 6))
    for a in range(6):
        R[mandel[a][0], mandel[a][1], a] = 1.0 / mandel_mults[a]
        R[mandel[a][1], mandel[a][0], a] = 1.0 / mandel_mults[a]

    return R


U = make_usym().reshape(9, 6)


def make_sym():
    """
    Make the tensor that goes from tensors to Mandel
    """
    R = np.zeros((6, 3, 3))
    for a in range(6):
        R[a, mandel[a][0], mandel[a][1]] = mandel_mults[a]

    return R


S = make_sym().reshape(6, 9)


def ms2ts(C):
    """
    Convert a Mandel notation stiffness matrix to a full stiffness tensor.
    """
    Ct = np.zeros((3, 3, 3, 3))
    for a in range(6):
        for b in range(6):
            ind_a = itertools.permutations(mandel[a], r=2)
            ind_b = itertools.permutations(mandel[b], r=2)
            ma = mandel_mults[a]
            mb = mandel_mults[b]
            indexes = tuple(ai + bi for ai, bi in itertools.product(ind_a, ind_b))
            for ind in indexes:
                Ct[ind] = C[a, b] / (ma * mb)

    return Ct


def ms2ts_faster(C):
    """
    Get cute with symmetry
    """
    # pylint: disable=too-many-function-args
    return np.dot(M2T, C.flatten()).reshape(3, 3, 3, 3)


def ts2ms(C):
    """
    Convert a stiffness tensor into a Mandel notation stiffness matrix
    """
    Cv = np.zeros((6, 6))
    for i in range(6):
        for j in range(6):
            ma = mandel_mults[i]
            mb = mandel_mults[j]
            Cv[i, j] = C[mandel[i] + mandel[j]] * ma * mb

    return Cv


def sym(A):
    """
    Take a symmetric matrix to the Mandel convention vector.
    """
    return np.array(
        [
            A[0, 0],
            A[1, 1],
            A[2, 2],
            np.sqrt(2) * A[1, 2],
            np.sqrt(2) * A[0, 2],
            np.sqrt(2) * A[0, 1],
        ]
    )


def sym_faster(A):
    """
    Take a symmetric tensor to the Mandel convention with multiplication
    """
    return np.dot(S, A.flatten())


def usym(v):
    """
    Take a Mandel symmetric vector to the full matrix.
    """
    return np.array(
        [
            [v[0], v[5] / np.sqrt(2), v[4] / np.sqrt(2)],
            [v[5] / np.sqrt(2), v[1], v[3] / np.sqrt(2)],
            [v[4] / np.sqrt(2), v[3] / np.sqrt(2), v[2]],
        ]
    )


def usym_faster(v):
    """
    Take a tensor to Mandel using multiplication
    """
    return np.dot(U, v).reshape(3, 3)



# pylint: disable=no-member, wrong-import-position
"""
    Thermal-fluid specific material classes
"""
import xml.etree.ElementTree as ET

from jax.config import config

config.update("jax_enable_x64", True)
import jax.numpy as jnp

from srlife import materials


class ThermalFluidMaterial:
    """Thermal fluid material properties

    The goal of this class is to provide film coefficient as a function of
    fluid velocity, temperature, and tube inner radius.

    The current implementation works from the
    - Heat capacity (constant pressure)
    - Density
    - Dynamic viscosity
    - Conductivity

    as a function of temperature (in K).
    """

    def __init__(
        self,
        film_min=1e-8,
        T_max=2000.0,
        T_min=0.0,
        laminar_cutoff=2e3,
        laminar_value=4.01,
    ):
        self.film_min = film_min
        self.T_max = T_max
        self.T_min = T_min
        self.laminar_cutoff = laminar_cutoff
        self.laminar_value = laminar_value

    def T_effective(self, T):
        """
        Cutoff T to avoid issues with the correlations
        """
        return jnp.maximum(jnp.minimum(T, self.T_max), self.T_min)

    @classmethod
    def load(cls, fname, modelname):
        """Load model from an XML file

        Parameters:
            fname:      XML filename
            modelname:  model name in the file
        """
        tag, mtype = materials.find_name(fname, modelname)
        data = materials.load_node(tag)[modelname]

        if mtype == "PolynomialThermalFluidMaterial":
            return PolynomialThermalFluidMaterial.load(data)
        else:
            raise ValueError("Unknown ThermalFluidMaterial type %s" % mtype)

    def save(self, fname, modelname):
        """Save model to an XML file

        Parameters:
            fname:      XML filename
            modelname:  model name in the file
        """
        root = ET.Element("models")

        materials.save_node(
            modelname, self.data(), root, attrib={"type": self.model_type}
        )
        tree = ET.ElementTree(element=root)
        tree.write(fname)

    def film_coefficient(self, T, u, r):
        """Film coefficient

        Calculate the film coefficient as a function of
        temperature, flow velocity, and tube diameter

        Parameters:
            T:      temperature, in K
            u:      flow velocity, in mm/hr
            r:      tube inner radius, in mm
        """
        nu = self.nusselt(T, u, r)

        return jnp.maximum(nu * self.k(T) / (2.0 * r), self.film_min)

    def reynolds(self, T, u, r):
        """Reynolds number

        Parameters:
            T:      temperature, in K
            u:      flow velocity, in mm/hr
            r:      tube inner radius, in mm
        """
        return self.rho(T) * u * 2.0 * r / self.mu(T)

    def prandtl(self, T):
        """Prandtl number

        Parameters:
            T:      temperature, in K
        """
        return self.cp(T) * self.mu(T) / self.k(T)

    def nusselt(self, T, u, r):
        """Nusselt number

        This is really where the work gets done in calculating the film
        coefficient and different correlations are  possible.

        Right now we use the Gnielinski correlation for turbulent flow

        Parameters:
            T:      temperature, in K
            u:      flow velocity, in m/s
            r:      tube inner radius, in mm
        """
        re = self.reynolds(self.T_effective(T), u, r)
        pr = self.prandtl(self.T_effective(T))
        f = (0.79 * jnp.log(re) - 1.64) ** -2.0

        turbulent = ((f / 8.0) * (re - 1000.0) * pr) / (
            1.0 + 12.7 * (f / 8.0) ** 0.5 * (pr ** (2.0 / 3.0) - 1.0)
        )

        turbulent.at[re < self.laminar_cutoff].set(self.laminar_value)

        return turbulent


class PolynomialThermalFluidMaterial(ThermalFluidMaterial):
    """Thermal-fluid relations correlated with polynomials

    The polynomial definitions use C, the class handles converting K -> C

    Parameters:
        cp_poly (array): coefficients for the heat capacity, in numpy order
        rho_poly (array): coefficients for the density, in numpy order
        mu_poly (array): coefficients for the viscosity, in numpy order
        k_poly (array): coefficients for the conductivity, in numpy order
        film_min (Optional[float]): lower cutoff on the value of the film coefficient
        T_max (Optional[float]): upper cutoff on temperature used in finding values
        T_min (Optional[float]): lower cutoff on temperature used in finding values
        laminar_cutoff (Optional[float]): Reynolds number upper limit for laminar flow
        laminar_value (Optional[float]): Nusslet number to use in laminar flow
    """

    def __init__(self, cp_poly, rho_poly, mu_poly, k_poly, **kwargs):
        super().__init__(**kwargs)
        self.cp_poly = jnp.asarray(cp_poly)
        self.rho_poly = jnp.asarray(rho_poly)
        self.mu_poly = jnp.asarray(mu_poly)
        self.k_poly = jnp.asarray(k_poly)

        self.model_type = "PolynomialThermalFluidMaterial"

    def data(self):
        """
        Reduce the class down to a dictionary of strings for
        serialization
        """
        return {
            "cp_poly": materials.string_array(self.cp_poly),
            "rho_poly": materials.string_array(self.rho_poly),
            "mu_poly": materials.string_array(self.mu_poly),
            "k_poly": materials.string_array(self.k_poly),
            "film_min": str(self.film_min),
            "T_max": str(self.T_max),
            "T_min": str(self.T_min),
            "laminar_cutoff": str(self.laminar_cutoff),
            "laminar_value": str(self.laminar_value),
        }

    @classmethod
    def load(cls, values):
        """Create from a dictionary

        Parameters:
            values:     dictionary of values
        """
        # Assemble kwargs
        kwargs = {
            k: float(values[k])
            for k in ["film_min", "T_max", "T_min", "laminar_cutoff", "laminar_value"]
            if k in values
        }

        return cls(
            materials.destring_array(values["cp_poly"]),
            materials.destring_array(values["rho_poly"]),
            materials.destring_array(values["mu_poly"]),
            materials.destring_array(values["k_poly"]),
            **kwargs
        )

    def cp(self, T):
        """Heat capacity as a function of temperature in K

        Parameters:
            T:      temperature, in K
        """
        return jnp.polyval(self.cp_poly, T)

    def rho(self, T):
        """Density as a function of temperature in K

        Parameters:
            T:      temperature, in K
        """
        return jnp.polyval(self.rho_poly, self.T_effective(T))

    def mu(self, T):
        """Dynamic viscosity, as a function of temperature in K

        Parameters:
            T:  temperature, in K
        """
        return jnp.polyval(self.mu_poly, T)

    def k(self, T):
        """Conductivity, as a function of temperature in K

        Parameters:
            T:  temperature, in K
        """
        return jnp.polyval(self.k_poly, T)



import numpy as np
import numpy.linalg as la
import networkx as nx


class Network(nx.MultiGraph):
    """
    Simplified version of panel network where everything is an elastic spring

    Boundary conditions are always fixed on the bottom and free on the top
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(args) == 0:
            self.add_node("top", distance=0)
            self.cpanel = 0

    def add_panel(self, k):
        """
        Add a panel to the network with stiffness k

        Parameters:
          k:      either a float stiffness or "rigid" or "disconnect"

        Returns:
          id of the new panel
        """
        name = self._panel_name(self.cpanel)
        self.cpanel += 1
        self.add_node(name, distance=1)
        self.add_edge("top", name, stiffness=k)

        return self.cpanel - 1

    def add_tube(self, panel, kconnect, k):
        """
        Add a tube to a panel

        Parameters:
          panel:      panel id
          kconnect:   either a float stiffness or "rigid" or "disconnect"
          k:          float stiffness
        """
        if panel >= self.cpanel:
            raise ValueError("Panel %i does not exist!" % panel)

        name = self._tube_name(panel, self._next_tube(panel))
        self.add_node(name + "_top", distance=2)
        self.add_node(name + "_bottom", distance=3)

        self.add_edge(self._panel_name(panel), name + "_top", stiffness=kconnect)
        self.add_edge(name + "_top", name + "_bottom", stiffness=k)

    def simplify_network(self):
        """
        Simplify the tube network and return a list of subproblems
        """
        # Remove all rigid nodes
        self._remove_rigid()

        # Split at all disconnects
        return self._split_disconnect()

    def solve_network(self, tforce):
        """
        Impose the standard BCs and solve for the displacements at each node

        Parameters:
          tforce:     thermal force on the bars
        """
        n = len(self)
        K = np.zeros((n, n))
        d = np.zeros((n,))
        f = np.zeros((n,))

        # Make the dof maps
        self.backmap = {}
        for i, name in enumerate(self.nodes()):
            self.backmap[name] = i

        # Fix all the ones with "_bottom" in the name to zero
        fixed = []
        for i, name in enumerate(self.nodes()):
            if name[:4] == "tube" and name.split("_")[1] == "bottom":
                fixed.append(i)
        free = sorted(list(set(list(range(n))) - set(fixed)))

        # Impose the thermal force and the stiffness matrix
        for a, b, data in self.edges(data=True):
            dofs = [self.backmap[a], self.backmap[b]]
            s = np.sign(self.nodes[b]["distance"] - self.nodes[a]["distance"])
            K[np.ix_(dofs, dofs)] += data["stiffness"] * np.array([[1, -1], [-1, 1]])

            if a[:4] == "tube" and b[:4] == "tube":
                f[dofs] += tforce * np.array([1, -1]) * s

        # Solve
        fprime = f[free] - np.dot(K[np.ix_(free, fixed)], d[fixed])
        Kprime = K[np.ix_(free, free)]
        dprime = la.solve(Kprime, fprime)
        d[free] = dprime

        # Backsub the displacements
        for i, (name, data) in enumerate(self.nodes(data=True)):
            data["displacement"] = d[i]

    def _remove_rigid(self):
        """
        Remove all rigid connections by deleting the edge and joining the
        nodes
        """
        while True:
            for a, b, key, data in self.edges(data=True, keys=True):
                if data["stiffness"] == "rigid":
                    self.remove_edge(a, b, key=key)
                    # Need to keep the one marked "tube"
                    if a[:4] == "tube":
                        ng = nx.contracted_nodes(self, a, b)
                    else:
                        ng = nx.contracted_nodes(self, b, a)
                    self.clear()
                    self.update(ng)
                    break
            else:
                break

    def _split_disconnect(self):
        """
        Split the graph at all disconnects
        """
        # First just split
        while True:
            for i, j, key, data in self.edges(keys=True, data=True):
                if data["stiffness"] == "disconnect":
                    self.remove_edge(i, j, key=key)
                    break
            else:
                break

        # Setup the new structures
        base = [
            Network(self.subgraph(c))
            for c in nx.connected_components(self)
            if len(c) > 1
        ]
        nstructs = []
        for b in base:
            ns = [n[:4] for n in b.nodes()]
            if "tube" in ns:
                nstructs.append(b)

        return nstructs

    def _panel_name(self, k):
        """
        Turn an integer into a panel id

        Parameters:
          k:      panel integer id
        """
        return "panel-%i" % k

    def _tube_name(self, panel, tube):
        """
        Turn integer panel and tube ids into a base tube string id
        """
        return "tube-%i-%i" % (panel, tube)

    def _next_tube(self, panel):
        """
        Figure out what the next tube id is for a panel
        """
        start = 0
        for n in self.neighbors(self._panel_name(panel)):
            if n != "top":
                start += 1

        return start



# This file lightly adapted from the example CMake build provided in the pybind11 distribution
# located at https://github.com/pybind/cmake_example/blob/master/setup.py
# That file is released under an open source license, contained in the pybind11 subdirectory
# of this package.

import os
import sys
import platform

from setuptools import setup, find_packages

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding = 'utf-8') as f:
  long_description = f.read()

setup (
    # Name of the project
    name = 'srlife',
    # Version
    version = '2.0.0',
    # One line-description
    description = "Evaluate the structural life of a solar receiver",
    # README
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    # Project webpage
    url='https://github.com/Argonne-National-Laboratory/srlife',
    # Author
    author='Argonne National Laboratory',
    # email
    author_email = 'messner@anl.gov',
    # Various things for pypi
    classifiers=[
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Operating System :: OS Independent'
      ],
    # Which version of python is needed
    python_requires='>=3.6',
    # Keywords to help find
    keywords='materials structures modeling',
    # Definitely not zip safe
    zip_safe=True,
    # Get the python files
    packages=find_packages(),
    # Python dependencies
    install_requires=[
      'numpy >= 1.18',
      'scipy',
      'h5py',
      'vtk ; platform_system!="Darwin"', 
      'nose', 
      'matplotlib',
      'pylint', 
      'neml', 
      'scikit-fem == 3.2.0', 
      'meshio', 
      'networkx', 
      'tqdm', 
      'multiprocess', 
      'dill'
      ],
    include_package_data=True,
    package_data={'': ['data/damage/*.xml', 'data/deformation/*.xml', 
      'data/fluid/*.xml', 'data/thermal/*.xml']},
)
